# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from connexion import NoContent
from flask import current_app, g, request
from marshmallow import ValidationError

from airflow.api_connexion import security
from airflow.api_connexion.exceptions import AlreadyExists, BadRequest, NotFound
from airflow.api_connexion.parameters import check_limit, format_datetime, format_parameters
from airflow.api_connexion.schemas.dag_run_schema import (
    DAGRunCollection,
    dagrun_collection_schema,
    dagrun_schema,
    dagruns_batch_form_schema,
)
from airflow.models import DagModel, DagRun
from airflow.security import permissions
from airflow.utils.session import provide_session
from airflow.utils.types import DagRunType


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
@provide_session
def delete_dag_run(dag_id, dag_run_id, session):
    """Delete a DAG Run"""
    if session.query(DagRun).filter(DagRun.dag_id == dag_id, DagRun.run_id == dag_run_id).delete() == 0:
        raise NotFound(detail=f"DAGRun with DAG ID: '{dag_id}' and DagRun ID: '{dag_run_id}' not found")
    return NoContent, 204


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
    ]
)
@provide_session
def get_dag_run(dag_id, dag_run_id, session):
    """Get a DAG Run."""
    dag_run = session.query(DagRun).filter(DagRun.dag_id == dag_id, DagRun.run_id == dag_run_id).one_or_none()
    if dag_run is None:
        raise NotFound(
            "DAGRun not found",
            detail=f"DAGRun with DAG ID: '{dag_id}' and DagRun ID: '{dag_run_id}' not found",
        )
    return dagrun_schema.dump(dag_run)


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
    ]
)
@format_parameters(
    {
        'start_date_gte': format_datetime,
        'start_date_lte': format_datetime,
        'execution_date_gte': format_datetime,
        'execution_date_lte': format_datetime,
        'end_date_gte': format_datetime,
        'end_date_lte': format_datetime,
        'limit': check_limit,
    }
)
@provide_session
def get_dag_runs(
    session,
    dag_id,
    start_date_gte=None,
    start_date_lte=None,
    execution_date_gte=None,
    execution_date_lte=None,
    end_date_gte=None,
    end_date_lte=None,
    offset=None,
    limit=None,
):
    """Get all DAG Runs."""
    query = session.query(DagRun)

    #  This endpoint allows specifying ~ as the dag_id to retrieve DAG Runs for all DAGs.
    if dag_id == "~":
        appbuilder = current_app.appbuilder
        query = query.filter(DagRun.dag_id.in_(appbuilder.sm.get_readable_dag_ids(g.user)))
    else:
        query = query.filter(DagRun.dag_id == dag_id)

    dag_run, total_entries = _fetch_dag_runs(
        query,
        end_date_gte,
        end_date_lte,
        execution_date_gte,
        execution_date_lte,
        start_date_gte,
        start_date_lte,
        limit,
        offset,
    )

    return dagrun_collection_schema.dump(DAGRunCollection(dag_runs=dag_run, total_entries=total_entries))


def _fetch_dag_runs(
    query,
    end_date_gte,
    end_date_lte,
    execution_date_gte,
    execution_date_lte,
    start_date_gte,
    start_date_lte,
    limit,
    offset,
):
    query = _apply_date_filters_to_query(
        query,
        end_date_gte,
        end_date_lte,
        execution_date_gte,
        execution_date_lte,
        start_date_gte,
        start_date_lte,
    )
    # Count items
    total_entries = query.count()
    # apply offset and limit
    dag_run = query.order_by(DagRun.id).offset(offset).limit(limit).all()
    return dag_run, total_entries


def _apply_date_filters_to_query(
    query, end_date_gte, end_date_lte, execution_date_gte, execution_date_lte, start_date_gte, start_date_lte
):
    # filter start date
    if start_date_gte:
        query = query.filter(DagRun.start_date >= start_date_gte)
    if start_date_lte:
        query = query.filter(DagRun.start_date <= start_date_lte)
    # filter execution date
    if execution_date_gte:
        query = query.filter(DagRun.execution_date >= execution_date_gte)
    if execution_date_lte:
        query = query.filter(DagRun.execution_date <= execution_date_lte)
    # filter end date
    if end_date_gte:
        query = query.filter(DagRun.end_date >= end_date_gte)
    if end_date_lte:
        query = query.filter(DagRun.end_date <= end_date_lte)
    return query


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG_RUN),
    ]
)
@provide_session
def get_dag_runs_batch(session):
    """Get list of DAG Runs"""
    body = request.get_json()
    try:
        data = dagruns_batch_form_schema.load(body)
    except ValidationError as err:
        raise BadRequest(detail=str(err.messages))

    appbuilder = current_app.appbuilder
    readable_dag_ids = appbuilder.sm.get_readable_dag_ids(g.user)
    query = session.query(DagRun)
    if data.get("dag_ids"):
        dag_ids = set(data["dag_ids"]) & set(readable_dag_ids)
        query = query.filter(DagRun.dag_id.in_(dag_ids))
    else:
        query = query.filter(DagRun.dag_id.in_(readable_dag_ids))

    dag_runs, total_entries = _fetch_dag_runs(
        query,
        data["end_date_gte"],
        data["end_date_lte"],
        data["execution_date_gte"],
        data["execution_date_lte"],
        data["start_date_gte"],
        data["start_date_lte"],
        data["page_limit"],
        data["page_offset"],
    )

    return dagrun_collection_schema.dump(DAGRunCollection(dag_runs=dag_runs, total_entries=total_entries))


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_CREATE, permissions.RESOURCE_DAG_RUN),
    ]
)
@provide_session
def post_dag_run(dag_id, session):
    """Trigger a DAG."""
    if not session.query(DagModel).filter(DagModel.dag_id == dag_id).first():
        raise NotFound(title="DAG not found", detail=f"DAG with dag_id: '{dag_id}' not found")
    try:
        post_body = dagrun_schema.load(request.json, session=session)
    except ValidationError as err:
        raise BadRequest(detail=str(err))
    dagrun_instance = (
        session.query(DagRun).filter(DagRun.dag_id == dag_id, DagRun.run_id == post_body["run_id"]).first()
    )
    if not dagrun_instance:
        dag_run = DagRun(dag_id=dag_id, run_type=DagRunType.MANUAL, **post_body)
        session.add(dag_run)
        session.commit()
        return dagrun_schema.dump(dag_run)
    raise AlreadyExists(
        detail=f"DAGRun with DAG ID: '{dag_id}' and DAGRun ID: '{post_body['run_id']}' already exists"
    )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import zipfile


def make_dir_zipfile(source_dir, output_filename):
    relroot = os.path.abspath(os.path.join(source_dir, os.pardir))
    with zipfile.ZipFile(output_filename, "w", zipfile.ZIP_DEFLATED) as zip:
        for root, dirs, files in os.walk(source_dir):
            # add directory (needed for empty dirs)
            zip.write(root, os.path.relpath(root, relroot))
            for file in files:
                filename = os.path.join(root, file)
                if os.path.isfile(filename): # regular files only
                    arcname = os.path.join(os.path.relpath(root, relroot), file)
                    zip.write(filename, arcname)


def make_file_zipfile(source_file, output_filename):
    with zipfile.ZipFile(output_filename, "w", zipfile.ZIP_DEFLATED) as zip_ref:
        zip_ref.write(source_file, os.path.basename(source_file))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Optional, List

import grpc

from ai_flow.model_center.entity.model_version import ModelVersion
from ai_flow.model_center.entity.model_version_detail import ModelVersionDetail
from ai_flow.model_center.entity.model_version_stage import ModelVersionStage
from ai_flow.model_center.entity.registered_model import RegisteredModel
from ai_flow.model_center.entity.registered_model_detail import RegisteredModelDetail
from ai_flow.protobuf import model_center_service_pb2_grpc
from ai_flow.protobuf.message_pb2 import ModelMetaParam, RegisteredModelParam, \
    ModelVersionParam, RegisteredModelDetail as ProtoModelDetail, RegisteredModelMeta, RegisteredModelMetas, \
    ModelVersionMeta
from ai_flow.protobuf.model_center_service_pb2 import CreateRegisteredModelRequest, \
    UpdateRegisteredModelRequest, DeleteRegisteredModelRequest, ListRegisteredModelsRequest, \
    GetRegisteredModelDetailRequest, CreateModelVersionRequest, UpdateModelVersionRequest, DeleteModelVersionRequest, \
    GetModelVersionDetailRequest
from ai_flow.endpoint.server import stringValue
from ai_flow.endpoint.client.base_client import BaseClient
from ai_flow.endpoint.server.util import _parse_response


class ModelCenterClient(BaseClient):
    def __init__(self, server_uri):
        super(ModelCenterClient, self).__init__(server_uri)
        channel = grpc.insecure_channel(server_uri)
        self.model_center_stub = model_center_service_pb2_grpc.ModelCenterServiceStub(channel)

    def create_registered_model(self, model_name, model_desc=None) -> Optional[RegisteredModelDetail]:
        """
        Create a new registered model from given type in Model Center.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param model_desc: (Optional) Description of registered model.

        :return: A single object of :py:class:`ai_flow.model_center.entity.RegisteredModelDetail` created in
        Model Center.
        """
        request = CreateRegisteredModelRequest(
            registered_model=RegisteredModelParam(model_name=stringValue(model_name),
                                                  model_desc=stringValue(model_desc)))
        response = self.model_center_stub.createRegisteredModel(request)
        return RegisteredModelDetail.from_proto(_parse_response(response, RegisteredModelMeta()))

    def update_registered_model(self, model_name, new_name=None,
                                model_desc=None) -> Optional[RegisteredModelDetail]:
        """
        Update metadata for RegisteredModel entity backend. Either ``model_name`` or ``model_desc``
        should be non-None. Backend raises exception if a registered model with given name does not exist.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param new_name: (Optional) New proposed name for the registered model.
        :param model_desc: (Optional) Description of registered model.

        :return: A single updated :py:class:`ai_flow.model_center.entity.RegisteredModelDetail` object.
        """
        request = UpdateRegisteredModelRequest(
            model_meta=ModelMetaParam(model_name=stringValue(model_name)),
            registered_model=RegisteredModelParam(model_name=stringValue(new_name),
                                                  model_desc=stringValue(model_desc)))
        response = self.model_center_stub.updateRegisteredModel(request)
        return RegisteredModelDetail.from_proto(_parse_response(response, RegisteredModelMeta()))

    def delete_registered_model(self, model_name) -> RegisteredModelDetail:
        """
        Delete registered model by model name in Model Center backend.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.

        :return: A single :py:class:`ai_flow.entities.model_registry.RegisteredModel` object.
        """
        request = DeleteRegisteredModelRequest(model_meta=ModelMetaParam(model_name=stringValue(model_name)))
        response = self.model_center_stub.deleteRegisteredModel(request)
        return RegisteredModel.from_resp_proto(_parse_response(response, ModelMetaParam()))

    def list_registered_models(self) -> List[RegisteredModelDetail]:
        """
        List of all registered models in Model Center backend.

        :return: List of :py:class:`ai_flow.model_center.entity.RegisteredModel` objects.
        """
        request = ListRegisteredModelsRequest()
        response = self.model_center_stub.listRegisteredModels(request)
        registered_model_details = []
        for registered_model_detail in _parse_response(response, RegisteredModelMetas()).registered_models:
            registered_model_details.append(RegisteredModelDetail.from_proto(registered_model_detail))
        return registered_model_details

    def get_registered_model_detail(self, model_name) -> Optional[RegisteredModelDetail]:
        """
        Get registered model detail filter by model name for Model Center.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.

        :return: A single :py:class:`ai_flow.entities.model_registry.RegisteredModelDetail` object.
        """
        request = GetRegisteredModelDetailRequest(model_meta=ModelMetaParam(model_name=stringValue(model_name)))
        response = self.model_center_stub.getRegisteredModelDetail(request)
        return RegisteredModelDetail.from_detail_proto(_parse_response(response, ProtoModelDetail()))

    def create_model_version(self, model_name, model_path, model_type=None, version_desc=None,
                             current_stage=ModelVersionStage.GENERATED) -> Optional[ModelVersionDetail]:
        """
        Create a new model version from given source and metric in Model Center.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param model_path: Source path where the AIFlow model is stored.
        :param model_type: (Optional) Type of AIFlow registered model option.
        :param version_desc: (Optional) Description of registered model version.
        :param current_stage: (Optional) Stage of registered model version.

        :return: A single object of :py:class:`ai_flow.model_center.entity.ModelVersionDetail` created in
        model repository.
        """
        request = CreateModelVersionRequest(model_meta=ModelMetaParam(model_name=stringValue(model_name)),
                                            model_version=ModelVersionParam(model_path=stringValue(model_path),
                                                                            model_type=stringValue(model_type),
                                                                            version_desc=stringValue(version_desc),
                                                                            current_stage=current_stage))
        response = self.model_center_stub.createModelVersion(request)
        return ModelVersionDetail.from_proto(_parse_response(response, ModelVersionMeta()))

    def update_model_version(self, model_name, model_version, model_path=None, model_type=None,
                             version_desc=None, current_stage=None) -> Optional[ModelVersionDetail]:
        """
        Update metadata for ModelVersion entity and metadata associated with a model version in backend.
        Either ``model_path`` or ``model_type`` or ``version_desc`` should be non-None.
        Backend raises exception if a registered model with given name does not exist.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param model_version: User-defined version of registered model.
        :param model_path: (Optional) Source path where the AIFlow model is stored.
        :param model_type: (Optional) Type of AIFlow registered model option.
        :param version_desc: (Optional) Description of registered model version.
        :param current_stage: (Optional) Current stage of registered model version.

        :return: A single updated :py:class:`ai_flow.model_center.entity.ModelVersionDetail` object.
        """
        request = UpdateModelVersionRequest(
            model_meta=ModelMetaParam(model_name=stringValue(model_name), model_version=stringValue(model_version)),
            model_version=ModelVersionParam(model_path=stringValue(model_path),
                                            model_type=stringValue(model_type),
                                            version_desc=stringValue(version_desc),
                                            current_stage=current_stage))
        response = self.model_center_stub.updateModelVersion(request)
        return ModelVersionDetail.from_proto(_parse_response(response, ModelVersionMeta()))

    def delete_model_version(self, model_name, model_version) -> ModelVersion:
        """
        Delete model version by model name and version in Model Center backend.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param model_version: User-defined version of registered model.

        :return: A single :py:class:`ai_flow.entities.model_registry.ModelVersion` object.
        """
        request = DeleteModelVersionRequest(model_meta=ModelMetaParam(model_name=stringValue(model_name),
                                                                      model_version=stringValue(model_version)))
        response = self.model_center_stub.deleteModelVersion(request)
        return ModelVersion.from_resp_proto(_parse_response(response, ModelMetaParam()))

    def get_model_version_detail(self, model_name, model_version) -> Optional[ModelVersionDetail]:
        """
        Get model version detail filter by model name and model version for Model Center.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param model_version: User-defined version of registered model.

        :return: A single :py:class:`ai_flow.entities.model_registry.ModelVersionDetail` object.
        """
        request = GetModelVersionDetailRequest(model_meta=ModelMetaParam(model_name=stringValue(model_name),
                                                                         model_version=stringValue(model_version)))
        response = self.model_center_stub.getModelVersionDetail(request)
        return ModelVersionDetail.from_proto(_parse_response(response, ModelVersionMeta()))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Add fractional seconds to mysql tables

Revision ID: 4addfa1236f1
Revises: f2ca10b85618
Create Date: 2016-09-11 13:39:18.592072

"""

from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = '4addfa1236f1'
down_revision = 'f2ca10b85618'
branch_labels = None
depends_on = None


def upgrade():  # noqa: D103
    conn = op.get_bind()  # pylint: disable=no-member
    if conn.dialect.name == "mysql":
        op.alter_column(table_name='dag', column_name='last_scheduler_run', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='dag', column_name='last_pickled', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='dag', column_name='last_expired', type_=mysql.DATETIME(fsp=6))

        op.alter_column(table_name='dag_pickle', column_name='created_dttm', type_=mysql.DATETIME(fsp=6))

        op.alter_column(table_name='dag_run', column_name='execution_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='dag_run', column_name='start_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='dag_run', column_name='end_date', type_=mysql.DATETIME(fsp=6))

        op.alter_column(table_name='import_error', column_name='timestamp', type_=mysql.DATETIME(fsp=6))

        op.alter_column(table_name='job', column_name='start_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='job', column_name='end_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='job', column_name='latest_heartbeat', type_=mysql.DATETIME(fsp=6))

        op.alter_column(table_name='log', column_name='dttm', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='log', column_name='execution_date', type_=mysql.DATETIME(fsp=6))

        op.alter_column(
            table_name='sla_miss', column_name='execution_date', type_=mysql.DATETIME(fsp=6), nullable=False
        )
        op.alter_column(table_name='sla_miss', column_name='timestamp', type_=mysql.DATETIME(fsp=6))

        op.alter_column(table_name='task_fail', column_name='execution_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='task_fail', column_name='start_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='task_fail', column_name='end_date', type_=mysql.DATETIME(fsp=6))

        op.alter_column(
            table_name='task_instance',
            column_name='execution_date',
            type_=mysql.DATETIME(fsp=6),
            nullable=False,
        )
        op.alter_column(table_name='task_instance', column_name='start_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='task_instance', column_name='end_date', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='task_instance', column_name='queued_dttm', type_=mysql.DATETIME(fsp=6))

        op.alter_column(table_name='xcom', column_name='timestamp', type_=mysql.DATETIME(fsp=6))
        op.alter_column(table_name='xcom', column_name='execution_date', type_=mysql.DATETIME(fsp=6))


def downgrade():  # noqa: D103
    conn = op.get_bind()  # pylint: disable=no-member
    if conn.dialect.name == "mysql":
        op.alter_column(table_name='dag', column_name='last_scheduler_run', type_=mysql.DATETIME())
        op.alter_column(table_name='dag', column_name='last_pickled', type_=mysql.DATETIME())
        op.alter_column(table_name='dag', column_name='last_expired', type_=mysql.DATETIME())

        op.alter_column(table_name='dag_pickle', column_name='created_dttm', type_=mysql.DATETIME())

        op.alter_column(table_name='dag_run', column_name='execution_date', type_=mysql.DATETIME())
        op.alter_column(table_name='dag_run', column_name='start_date', type_=mysql.DATETIME())
        op.alter_column(table_name='dag_run', column_name='end_date', type_=mysql.DATETIME())

        op.alter_column(table_name='import_error', column_name='timestamp', type_=mysql.DATETIME())

        op.alter_column(table_name='job', column_name='start_date', type_=mysql.DATETIME())
        op.alter_column(table_name='job', column_name='end_date', type_=mysql.DATETIME())
        op.alter_column(table_name='job', column_name='latest_heartbeat', type_=mysql.DATETIME())

        op.alter_column(table_name='log', column_name='dttm', type_=mysql.DATETIME())
        op.alter_column(table_name='log', column_name='execution_date', type_=mysql.DATETIME())

        op.alter_column(
            table_name='sla_miss', column_name='execution_date', type_=mysql.DATETIME(), nullable=False
        )
        op.alter_column(table_name='sla_miss', column_name='timestamp', type_=mysql.DATETIME())

        op.alter_column(table_name='task_fail', column_name='execution_date', type_=mysql.DATETIME())
        op.alter_column(table_name='task_fail', column_name='start_date', type_=mysql.DATETIME())
        op.alter_column(table_name='task_fail', column_name='end_date', type_=mysql.DATETIME())

        op.alter_column(
            table_name='task_instance', column_name='execution_date', type_=mysql.DATETIME(), nullable=False
        )
        op.alter_column(table_name='task_instance', column_name='start_date', type_=mysql.DATETIME())
        op.alter_column(table_name='task_instance', column_name='end_date', type_=mysql.DATETIME())
        op.alter_column(table_name='task_instance', column_name='queued_dttm', type_=mysql.DATETIME())

        op.alter_column(table_name='xcom', column_name='timestamp', type_=mysql.DATETIME())
        op.alter_column(table_name='xcom', column_name='execution_date', type_=mysql.DATETIME())



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import time
import unittest
from typing import List
from unittest import TestCase

from ai_flow.project.project_config import ProjectConfig
from notification_service.base_notification import EventWatcher

from ai_flow.common.properties import Properties
from ai_flow.common.status import Status
from ai_flow.meta.dataset_meta import DatasetMeta, DataType, Schema
from ai_flow.meta.job_meta import State
from ai_flow.meta.metric_meta import MetricType, MetricMeta, MetricSummary
from ai_flow.model_center.entity.model_version_stage import ModelVersionStage
from ai_flow.protobuf.message_pb2 import RESOURCE_ALREADY_EXISTS
from ai_flow.client.ai_flow_client import AIFlowClient
from ai_flow.endpoint.server.exception import AIFlowException
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.store.db.base_model import base
from ai_flow.test.store.test_sqlalchemy_store import _get_store

_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)
_PORT = '50051'

client = None
client1 = None
client2 = None


class AIFlowClientTestCases(object):

    """test dataset"""

    def test_save_dataset_get_dataset_by_id_and_name(self):
        dataset = client.register_dataset(name='dataset', data_format='csv', description='it is mq data',
                                          uri='mysql://',
                                          properties=Properties({'a': 'b'}), name_list=['a'],
                                          type_list=[DataType.INT32])
        dataset_id = client.get_dataset_by_id(2)
        self.assertIsNone(dataset_id)
        dataset_name = client.get_dataset_by_name('dataset')
        self.assertEqual('dataset', dataset.name)
        self.assertEqual('dataset', dataset_name.name)

    def test_save_dataset_with_catalog_by_id_and_name(self):
        client.register_dataset_with_catalog(name='dataset',
                                             catalog_name='my_hive', catalog_connection_uri='/path/to/conf',
                                             catalog_type='hive', catalog_database='my_db', catalog_table='my_table')
        dataset_id = client.get_dataset_by_id(2)
        self.assertIsNone(dataset_id)
        dataset_name = client.get_dataset_by_name('dataset')
        self.assertEqual('my_hive', dataset_name.catalog_name)
        self.assertEqual('hive', dataset_name.catalog_type)
        self.assertEqual('my_db', dataset_name.catalog_database)
        self.assertEqual('my_table', dataset_name.catalog_table)
        self.assertEqual('/path/to/conf', dataset_name.catalog_connection_uri)

    def test_double_register_dataset(self):
        dataset_1 = client.register_dataset(name='dataset', data_format='csv', description='it is mq data',
                                            uri='mysql://', properties=Properties({'a': 'b'}), name_list=['a'],
                                            type_list=[DataType.INT32])

        dataset_2 = client.register_dataset(name='dataset', data_format='csv', description='it is mq data',
                                            uri='mysql://', properties=Properties({'a': 'b'}), name_list=['a'],
                                            type_list=[DataType.INT32])
        self.assertEqual(dataset_1.uuid, dataset_2.uuid)
        self.assertEqual(dataset_1.schema.to_json_dict(), dataset_2.schema.to_json_dict())
        self.assertRaises(AIFlowException, client.register_dataset, name='dataset',
                          data_format='csv',
                          description='it is not mq data', uri='mysql://',
                          properties=Properties({'a': 'b'}), name_list=['a'], type_list=[DataType.INT32])

    def test_list_datasets(self):
        client.register_dataset(name='dataset_1', data_format='csv', description='it is mq data',
                                uri='mysql://', properties=Properties({'a': 'b'}), name_list=['a'],
                                type_list=[DataType.INT32])
        client.register_dataset(name='dataset_2', data_format='npz', description='it is',
                                uri='mysql://', properties=Properties({'a': 'b'}), name_list=['a'],
                                type_list=[DataType.INT32])
        response_list = client.list_datasets(5, 0)
        self.assertEqual(len(response_list), 2)
        self.assertEqual('dataset_1', response_list[0].name)
        self.assertEqual('dataset_2', response_list[1].name)

    def test_save_datasets_list_datasets(self):
        dataset_1 = DatasetMeta(name='dataset1',
                                data_format='csv',
                                create_time=None, update_time=1000,
                                properties=Properties({'a': 'b'}))
        schema = Schema(name_list=['a', 'b'],
                        type_list=[DataType.STRING, DataType.INT32])
        dataset_2 = DatasetMeta(name='dataset2',
                                data_format='csv',
                                create_time=None, update_time=1000,
                                properties=Properties({'a': 'b'}), schema=schema)
        response = client.register_datasets([dataset_1, dataset_2])
        self.assertEqual(len(response), 2)
        self.assertEqual(1, response[0].uuid)
        self.assertEqual(2, response[1].uuid)
        response_list = client.list_datasets(2, 0)
        self.assertEqual(2, len(response_list))
        self.assertEqual('dataset1', response_list[0].name)
        self.assertEqual('dataset2', response_list[1].name)

    def test_delete_dataset(self):
        dataset = client.register_dataset(name='dataset',
                                          data_format='csv',
                                          description='it is mq data',
                                          uri='mysql://',
                                          properties=Properties({'a': 'b'}), name_list=['a'],
                                          type_list=[DataType.INT32])
        self.assertEqual(Status.OK, client.delete_dataset_by_name(dataset.name))
        self.assertIsNone(client.get_dataset_by_name(dataset.name))
        self.assertIsNone(client.list_datasets(1, 0))

    def test_update_dataset(self):
        client.register_dataset(name='dataset', data_format='csv', description='it is mq data',
                                uri='mysql://',
                                properties=Properties({'a': 'b'}), name_list=['a'], type_list=[DataType.INT32])
        now = int(time.time() * 1000)
        update_dataset = client.update_dataset(dataset_name='dataset', data_format='npz',
                                               properties=Properties({'kafka': 'localhost:9092'}),
                                               name_list=['b'], type_list=[DataType.STRING])
        dataset = client.get_dataset_by_name('dataset')
        self.assertTrue(dataset.update_time >= now)
        self.assertEqual(dataset.schema.name_list, update_dataset.schema.name_list)
        self.assertEqual(dataset.schema.type_list, update_dataset.schema.type_list)
        update_dataset_1 = client.update_dataset(dataset_name='dataset', catalog_type='hive', catalog_name='my_hive',
                                                 catalog_database='my_db', catalog_table='my_table')
        self.assertEqual(update_dataset_1.catalog_type, 'hive')
        self.assertEqual(update_dataset_1.catalog_name, 'my_hive')
        self.assertEqual(update_dataset_1.catalog_database, 'my_db')
        self.assertEqual(update_dataset_1.catalog_table, 'my_table')

    """test project"""

    def test_save_project_get_project_by_id_and_name(self):
        response = client.register_project(name='project', uri='www.code.com',)
        project_id = client.get_project_by_id(response.uuid)
        project_name = client.get_project_by_name('project')
        self.assertEqual(project_id.name, 'project')
        self.assertEqual(project_name.name, 'project')
        print(project_id)

    def test_double_register_project(self):
        client.register_project(name='project', uri='www.code.com')
        client.register_project(name='project', uri='www.code.com')
        self.assertRaises(AIFlowException, client.register_project,
                          name='project', uri='www.code2.com')

    def test_list_project(self):
        response = client.register_project(name='project', uri='www.code.com')
        client.register_project(name='project1', uri='www.code.com')
        project_list = client.list_project(2, response.uuid - 1)
        self.assertEqual(2, len(project_list))
        self.assertEqual('project', project_list[0].name)
        self.assertEqual('project1', project_list[1].name)

    def test_delete_project_by_id(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model_relation(name='model', project_id=project.uuid)

        client.register_model_version_relation(version='1', model_id=model.uuid,
                                               project_snapshot_id=None)
        self.assertEqual(client.get_project_by_id(project.uuid).name, 'project')
        self.assertEqual(client.get_model_relation_by_id(model.uuid).name, 'model')
        self.assertEqual(client.get_model_version_relation_by_version('1', 1).version, '1')
        self.assertEqual(Status.OK, client.delete_project_by_id(project.uuid))
        self.assertIsNone(client.get_project_by_id(project.uuid))
        self.assertIsNone(client.get_model_relation_by_id(model.uuid))
        self.assertIsNone(client.get_model_version_relation_by_version('1', model.uuid))
        self.assertIsNone(client.list_project(1, 0))
        self.assertIsNone(client.list_model_relation(1, 0))
        self.assertIsNone(client.list_model_version_relation(1, 1, 0))

    def test_delete_project_by_name(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model_relation(name='model', project_id=project.uuid)

        client.register_model_version_relation(version='1', model_id=model.uuid,
                                               project_snapshot_id=None)
        self.assertEqual(client.get_project_by_id(project.uuid).name, 'project')
        self.assertEqual(client.get_model_relation_by_id(model.uuid).name, 'model')
        self.assertEqual(client.get_model_version_relation_by_version('1', 1).version, '1')
        self.assertEqual(Status.OK, client.delete_project_by_id(project.uuid))
        self.assertIsNone(client.get_project_by_name('project'))
        self.assertIsNone(client.get_model_relation_by_id(model.uuid))
        self.assertIsNone(client.get_model_version_relation_by_version('1', model.uuid))
        self.assertIsNone(client.list_project(1, 0))
        self.assertIsNone(client.list_model_relation(1, 0))
        self.assertIsNone(client.list_model_version_relation(1, 1, 0))

    def test_update_project(self):
        client.register_project(name='project', uri='www.code.com')
        update_project = client.update_project(project_name='project', uri='git@alibaba.com')
        project = client.get_project_by_name('project')
        self.assertEqual(update_project.uri, project.uri)

    """test workflow"""

    def test_save_workflow_get_workflow_by_id_and_name(self):
        project_response = client.register_project(name='project', uri='www.code.com')
        self.assertEqual(project_response.uuid, 1)
        response = client.register_workflow(name='workflow',
                                            project_id=project_response.uuid,
                                            properties=Properties({'a': 'b'}))
        self.assertEqual(response.uuid, 1)
        self.assertEqual(response.properties, Properties({'a': 'b'}))
        response_by_id = client.get_workflow_by_id(response.uuid)
        response_by_name = client.get_workflow_by_name(project_response.name, response.name)
        self.assertEqual('workflow', response_by_id.name)
        self.assertEqual('workflow', response_by_name.name)
        self.assertEqual(Properties({'a': 'b'}), response_by_id.properties)
        self.assertEqual(Properties({'a': 'b'}), response_by_name.properties)

    def test_double_register_workflow(self):
        project_response = client.register_project(name='project', uri='www.code.com')
        project_response2 = client.register_project(name='project2', uri='www.code.com')
        client.register_workflow(name='workflow', project_id=project_response.uuid)
        client.register_workflow(name='workflow', project_id=project_response2.uuid)
        self.assertRaises(AIFlowException, client.register_workflow, name='workflow',
                          project_id=project_response.uuid)

    def test_list_workflows(self):
        project_response = client.register_project(name='project', uri='www.code.com')
        client.register_workflow(name='workflow1', project_id=project_response.uuid)
        client.register_workflow(name='workflow2', project_id=project_response.uuid)
        response_list = client.list_workflows(project_response.name, 2, 0)
        self.assertEqual('workflow1', response_list[0].name)
        self.assertEqual('workflow2', response_list[1].name)

    def test_delete_workflow(self):
        project_response = client.register_project(name='project', uri='www.code.com')
        response = client.register_workflow(name='workflow',
                                            project_id=project_response.uuid,
                                            properties=Properties({'a': 'b'}))
        self.assertEqual(Status.OK, client.delete_workflow_by_name(project_name=project_response.name,
                                                                   workflow_name='workflow'))
        self.assertIsNone(client.get_workflow_by_id(response.uuid))

        response = client.register_workflow(name='workflow', project_id=project_response.uuid)
        self.assertEqual(Status.OK, client.delete_workflow_by_id(response.uuid))
        self.assertIsNone(client.get_workflow_by_id(response.uuid))

    def test_update_workflow(self):
        project_response = client.register_project(name='project', uri='www.code.com')
        response = client.register_workflow(name='workflow',
                                            project_id=project_response.uuid,
                                            properties=Properties({'a': 'b'}))

        updated_workflow = client.update_workflow(project_name=project_response.name,
                                                  workflow_name='workflow',
                                                  properties=Properties({'a': 'c'}))
        self.assertEqual(updated_workflow.properties, Properties({'a': 'c'}))

    """test model"""

    def test_model_api(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model(model_name='test_register_model1',
                                      model_desc='test register model1', project_id=project.uuid)
        self.assertIsNone(client.get_model_by_name('no'))
        self.assertIsNone(client.get_model_by_id(2))
        self.assertEqual(client.get_model_by_id(model.uuid).name, 'test_register_model1')
        self.assertEqual(client.get_model_by_name('test_register_model1').name, 'test_register_model1')
        self.assertEqual(client.get_model_by_name('test_register_model1').model_desc, 'test register model1')
        client.register_model(model_name='test_register_model2',
                              model_desc='test register model2', project_id=1)
        self.assertEqual(len(client.list_model_relation(10, 0)), 2)

        client.delete_model_by_id(model.uuid)
        client.delete_model_by_name('test_register_model2')
        self.assertIsNone(client.list_model_relation(10, 0))
        self.assertEqual(len(client.list_registered_models()), 0)

    def test_get_deployed_model_version(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model(model_name='test_register_model1',
                                      model_desc='test register model1', project_id=project.uuid)
        model_version = client.register_model_version(model=model.uuid, model_path='/path/to/your/model/version')
        deployed_model_version = client.get_deployed_model_version(model_name=model.name)
        self.assertIsNone(deployed_model_version)
        client.update_model_version(model_name=model.name, model_version=model_version.version,
                                    current_stage=ModelVersionStage.DEPLOYED)
        deployed_model_version = client.get_deployed_model_version(model_name=model.name)
        self.assertEqual(deployed_model_version.version, model_version.version)
        self.assertRaises(AIFlowException,
                          client.update_model_version, model_name=model.name, model_version=model_version.version,
                          current_stage=ModelVersionStage.DEPLOYED)

    def test_save_model_get_id_and_name(self):
        project = client.register_project(name='project', uri='www.code.com')
        response = client.register_model_relation(name='model', project_id=project.uuid)
        model_id = client.get_model_relation_by_id(response.uuid)
        model_name = client.get_model_relation_by_name('model')
        self.assertEqual(model_id.name, model_name.name)
        self.assertEqual(1, len(client.list_model_relation(2, response.uuid - 1)))
        print(model_id)

    def test_list_model(self):
        project = client.register_project(name='project', uri='www.code.com')
        client.register_model_relation(name='model', project_id=project.uuid)
        client.register_model_relation(name='model1', project_id=project.uuid)
        self.assertEqual(2, len(client.list_model_relation(2, 0)))
        self.assertEqual('model', client.list_model_relation(2, 0)[0].name)
        self.assertEqual('model1', client.list_model_relation(2, 0)[1].name)

    def test_delete_model_by_id(self):
        project = client.register_project(name='project', uri='www.code.com')
        model_relation = client.register_model_relation(name='model', project_id=project.uuid)
        client.register_model_version_relation(version='1', model_id=model_relation.uuid,
                                               project_snapshot_id=None)
        self.assertEqual(client.get_model_version_relation_by_version('1', model_relation.uuid).version, '1')
        self.assertEqual(client.get_model_relation_by_name('model').name, 'model')
        self.assertEqual(Status.OK, client.delete_model_relation_by_id(model_relation.uuid))
        self.assertIsNone(client.get_model_version_relation_by_version('1', model_relation.uuid))
        self.assertIsNone(client.get_model_relation_by_name('model'))

    def test_delete_model_by_name(self):
        project = client.register_project(name='project', uri='www.code.com')
        model_relation = client.register_model_relation(name='model', project_id=project.uuid)
        client.register_model_version_relation(version='1', model_id=model_relation.uuid,
                                               project_snapshot_id=None)
        self.assertEqual(client.get_model_version_relation_by_version('1', model_relation.uuid).version, '1')
        self.assertEqual(client.get_model_relation_by_name('model').name, 'model')
        self.assertEqual(Status.OK, client.delete_model_relation_by_name('model'))
        self.assertIsNone(client.get_model_version_relation_by_version('1', model_relation.uuid))
        self.assertIsNone(client.get_model_relation_by_name('model'))

    """test model version"""

    def test_model_version_api(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model(model_name='test_register_model',
                                      model_desc='test register model', project_id=project.uuid)
        self.assertIsNone(client.get_model_version_by_version('1', model.uuid))
        self.assertEqual(client.get_model_by_id(model.uuid).name, 'test_register_model')
        self.assertEqual(client.get_model_by_name('test_register_model').name, 'test_register_model')
        response = client.register_model_version(model=model.uuid,
                                                 project_snapshot_id=None,
                                                 model_path='fs://source1.pkl',
                                                 version_desc='test model version 1',
                                                 current_stage=ModelVersionStage.GENERATED)
        self.assertEqual(response.version, '1')
        model_version_meta = client.get_model_version_by_version(response.version, model.uuid)
        self.assertEqual(model_version_meta.version, '1')
        self.assertEqual(model_version_meta.model_path, 'fs://source1.pkl')
        self.assertIsNone(model_version_meta.model_type)
        self.assertEqual(model_version_meta.version_desc, 'test model version 1')
        response = client.update_model_version(model_name=model.name, model_version='1',
                                               current_stage=ModelVersionStage.DEPLOYED)
        self.assertEqual(response.current_stage, ModelVersionStage.DEPLOYED)
        response = client.get_deployed_model_version(model.name)
        self.assertEqual(response.version, '1')

        response = client.register_model_version(model=model.uuid,
                                                 project_snapshot_id=None,
                                                 model_path='fs://source2.pkl',
                                                 model_type='{"flavor.version":2}',
                                                 version_desc='test model version 2')
        self.assertEqual(response.version, '2')
        self.assertEqual(len(client.list_model_version_relation(1, 10, 0)), 2)

        client.delete_model_version_by_version(version='2', model_id=1)
        self.assertEqual(len(client.list_model_version_relation(1, 10, 0)), 1)
        # register model version with deleted model version name
        response = client.register_model_version(model=model.uuid,
                                                 project_snapshot_id=None,
                                                 model_path='fs://source1.pkl',
                                                 version_desc='test model version 1')
        self.assertEqual(response.version, '2')
        model_version_meta = client.get_model_version_by_version(response.version, model.uuid)
        self.assertEqual(model_version_meta.version, '2')
        self.assertEqual(model_version_meta.model_path, 'fs://source1.pkl')
        self.assertIsNone(model_version_meta.model_type)
        self.assertEqual(model_version_meta.version_desc, 'test model version 1')

    def test_get_latest_model_version(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model(model_name='test_register_model',
                                      model_desc='test register model', project_id=project.uuid)
        response_1 = client.register_model_version(model=model.uuid,
                                                   project_snapshot_id=None,
                                                   model_path='fs://source1.pkl',
                                                   version_desc='test model version 1',
                                                   current_stage=ModelVersionStage.GENERATED)
        new_generated_model_version_1 = client.get_latest_generated_model_version(model.name)
        new_validated_model_version_1 = client.get_latest_validated_model_version(model.name)
        self.assertIsNone(new_validated_model_version_1)
        self.assertEqual(response_1.version, new_generated_model_version_1.version)
        client.update_model_version(model_name=model.name, model_version=response_1.version,
                                    current_stage=ModelVersionStage.VALIDATED)
        new_validated_model_version_2 = client.get_latest_validated_model_version(model.name)
        self.assertEqual(new_validated_model_version_2.version, response_1.version)
        response_2 = client.register_model_version(model=model.uuid,
                                                   project_snapshot_id=None,
                                                   model_path='fs://source1.pkl',
                                                   version_desc='test model version 1',
                                                   current_stage=ModelVersionStage.GENERATED)
        new_generated_model_version_2 = client.get_latest_generated_model_version(model.name)
        client.update_model_version(model_name=model.name, model_version=response_2.version,
                                    current_stage=ModelVersionStage.VALIDATED)
        new_validated_model_version_2 = client.get_latest_validated_model_version(model.name)
        self.assertEqual(new_validated_model_version_2.version, response_2.version)
        self.assertEqual(response_2.version, new_generated_model_version_2.version)

    def test_save_model_version_get_by_version(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model_relation(name='model', project_id=project.uuid)
        response = client.register_model_version_relation(version='1', model_id=model.uuid,
                                                          project_snapshot_id=None)
        self.assertEqual(response.version, '1')
        self.assertEqual(client.get_model_version_relation_by_version(response.version, model.uuid).version, '1')
        self.assertEqual(len(client.list_model_version_relation(model.uuid, 2, 0)), 1)
        print(client.get_model_version_relation_by_version(response.version, model.uuid))

    def test_list_model_version(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model_relation(name='model', project_id=project.uuid)
        client.register_model_version_relation(version='1', model_id=model.uuid,
                                               project_snapshot_id=None)
        client.register_model_version_relation(version='2', model_id=model.uuid,
                                               project_snapshot_id=None)
        self.assertEqual(len(client.list_model_version_relation(1, 2, 0)), 2)
        self.assertEqual(client.list_model_version_relation(1, 2, 0)[0].version, '1')
        self.assertEqual(client.list_model_version_relation(1, 2, 0)[1].version, '2')

    def test_delete_model_version_by_version(self):
        project = client.register_project(name='project', uri='www.code.com')
        model = client.register_model_relation(name='model', project_id=project.uuid)
        client.register_model_version_relation(version='1', model_id=model.uuid,
                                               project_snapshot_id=None)
        self.assertEqual(client.get_model_version_relation_by_version('1', model.uuid).version, '1')
        client.delete_model_version_relation_by_version('1', model.uuid)
        self.assertIsNone(client.get_model_version_relation_by_version('1', model.uuid))

        """test artifact"""

    def test_save_artifact_get_artifact_by_id_and_name(self):
        artifact = client.register_artifact(name='artifact', artifact_type='json', uri='./artifact.json')
        artifact_id = client.get_artifact_by_id(artifact.uuid)
        artifact_name = client.get_artifact_by_name(artifact.name)
        self.assertEqual(artifact.artifact_type, artifact_id.artifact_type)
        self.assertEqual('artifact', artifact_name.name)

    def test_double_save_artifact(self):
        artifact_1 = client.register_artifact(name='artifact', artifact_type='json', uri='./artifact.json')
        artifact_2 = client.register_artifact(name='artifact', artifact_type='json', uri='./artifact.json')
        self.assertEqual(artifact_1.to_json_dict(), artifact_2.to_json_dict())
        self.assertRaises(AIFlowException, client.register_artifact, name='artifact', artifact_type='json',
                          uri='./artifact.json', description='whatever')

    def test_save_artifact_list_artifact(self):
        client.register_artifact(name='artifact', artifact_type='json', uri='./artifact.json')
        client.register_artifact(name='artifact_1', artifact_type='json', uri='./artifact.json')
        self.assertEqual(2, len(client.list_artifact(2, 0)))

    def test_delete_artifact_by_id_and_name(self):
        client.register_artifact(name='artifact', artifact_type='json', uri='./artifact.json')
        client.register_artifact(name='artifact_1', artifact_type='json', uri='./artifact.json')
        self.assertIsNotNone(client.get_artifact_by_id(1))
        self.assertIsNotNone(client.get_artifact_by_name('artifact_1'))
        self.assertEqual(Status.OK, client.delete_artifact_by_id(1))
        self.assertEqual(Status.OK, client.delete_artifact_by_name('artifact_1'))
        self.assertEqual(Status.ERROR, client.delete_artifact_by_name('no artifact'))
        self.assertIsNone(client.get_artifact_by_id(1))
        self.assertIsNone(client.get_artifact_by_name('artifact_1'))

    def test_update_artifact(self):
        client.register_artifact(name='artifact', artifact_type='json', uri='./artifact.json')
        artifact = client.update_artifact(artifact_name='artifact', artifact_type='csv', uri='../..')
        artifact_id = client.get_artifact_by_id(artifact.uuid)
        self.assertEqual(artifact_id.artifact_type, 'csv')
        self.assertIsNotNone(artifact_id.update_time)
        self.assertEqual(artifact_id.uri, '../..')

    def test_create_registered_model(self):
        model_name = 'test_create_registered_model'
        model_desc = 'test create registered model'
        response = client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_desc, model_desc)

        with self.assertRaises(AIFlowException) as exception_context:
            client.create_registered_model(model_name=model_name)
        assert exception_context.exception.error_code == str(RESOURCE_ALREADY_EXISTS)

    def test_double_register_model(self):
        model_name = 'test_create_registered_model'
        model_desc = 'test create registered model'
        client.create_registered_model(model_name=model_name, model_desc=model_desc)
        client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertRaises(AIFlowException, client.create_registered_model, model_name=model_name,
                          model_desc='')
        project = client.register_project(name='project')
        client.register_model(model_name=model_name, project_id=project.uuid,
                              model_desc=model_desc)
        client.register_model(model_name=model_name, project_id=project.uuid,
                              model_desc=model_desc)
        self.assertRaises(AIFlowException, client.register_model, model_name=model_name,
                          project_id=project.uuid,
                          model_desc='')

    def test_update_registered_model(self):
        model_name1 = 'test_update_registered_model1'
        model_desc1 = 'test update registered model1'
        response = client.create_registered_model(model_name=model_name1,
                                                  model_desc=model_desc1)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name1)

        model_name2 = 'test_update_registered_model2'
        model_desc2 = 'test update registered model2'
        response = client.update_registered_model(model_name=model_name1, new_name=model_name2,
                                                  model_desc=model_desc2)
        self.assertEqual(response.model_name, model_name2)
        self.assertEqual(response.model_desc, model_desc2)

    def test_delete_registered_model(self):
        model_name = 'test_delete_registered_model'
        model_desc = 'test delete registered model'
        response = client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)

        client.delete_registered_model(model_name=model_name)
        response = client.get_registered_model_detail(model_name=model_name)
        self.assertIsNone(response)

    def test_list_registered_model(self):
        model_name1 = 'test_list_registered_model1'
        model_desc1 = 'test list registered model1'
        response = client.create_registered_model(model_name=model_name1,
                                                  model_desc=model_desc1)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name1)

        model_name2 = 'test_list_registered_model2'
        model_desc2 = 'test list registered model2'
        response = client.create_registered_model(model_name=model_name2,
                                                  model_desc=model_desc2)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name2)

        response = client.list_registered_models()
        self.assertEqual(len(response), 2)
        self.assertEqual(response[0].model_name, model_name1)
        self.assertEqual(response[1].model_name, model_name2)

    def test_get_registered_model_detail(self):
        model_name = 'test_get_registered_model_detail'
        model_desc = 'test get registered model detail'
        response = client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)

        response = client.get_registered_model_detail(model_name=model_name)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_desc, model_desc)

        model_path1 = 'fs://source1.pkl'
        model_type1 = '{"flavor.version":1}'
        version_desc1 = 'test get registered model detail1'
        response = client.create_model_version(model_name=model_name, model_path=model_path1,
                                               model_type=model_type1,
                                               version_desc=version_desc1)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '1')
        self.assertEqual(response.model_path, model_path1)
        self.assertEqual(response.model_type, model_type1)
        self.assertEqual(response.version_desc, version_desc1)

        response = client.get_registered_model_detail(model_name=model_name)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_desc, model_desc)
        model_version = response.latest_model_version
        self.assertEqual(model_version.model_version, '1')
        self.assertEqual(model_version.model_path, model_path1)
        self.assertEqual(model_version.model_type, model_type1)
        self.assertEqual(model_version.version_desc, version_desc1)

        model_path2 = 'fs://source2.pkl'
        model_type2 = '{"flavor.version":2}'
        version_desc2 = 'test get registered model detail2'
        response = client.create_model_version(model_name=model_name, model_path=model_path2,
                                               model_type=model_type2,
                                               version_desc=version_desc2)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '2')
        self.assertEqual(response.model_path, model_path2)
        self.assertEqual(response.model_type, model_type2)
        self.assertEqual(response.version_desc, version_desc2)

        response = client.get_registered_model_detail(model_name=model_name)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_desc, model_desc)
        model_version = response.latest_model_version
        self.assertEqual(model_version.model_version, '2')
        self.assertEqual(model_version.model_path, model_path2)
        self.assertEqual(model_version.model_type, model_type2)
        self.assertEqual(model_version.version_desc, version_desc2)

    def test_create_model_version(self):
        model_name = 'test_create_model_version'
        model_desc = 'test create model version'
        response = client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)

        model_path1 = 'fs://source1.pkl'
        model_type1 = '{"flavor.version":1}'
        version_desc1 = 'test create model version1'
        response = client.create_model_version(model_name=model_name, model_path=model_path1,
                                               model_type=model_type1,
                                               version_desc=version_desc1)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '1')
        self.assertEqual(response.model_path, model_path1)
        self.assertEqual(response.model_type, model_type1)
        self.assertEqual(response.version_desc, version_desc1)

        model_path2 = 'fs://source2.pkl'
        model_type2 = '{"flavor.version":2}'
        version_desc2 = 'test create model version2'
        response = client.create_model_version(model_name=model_name, model_path=model_path2,
                                               model_type=model_type2,
                                               version_desc=version_desc2)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '2')
        self.assertEqual(response.model_path, model_path2)
        self.assertEqual(response.model_type, model_type2)
        self.assertEqual(response.version_desc, version_desc2)

    def test_update_model_version(self):
        model_name = 'test_update_model_version'
        model_desc = 'test update model version'
        response = client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)

        model_path1 = 'fs://source1.pkl'
        model_type1 = '{"flavor.version":1}'
        version_desc1 = 'test update model version1'
        version_stage1 = ModelVersionStage.GENERATED
        response = client.create_model_version(model_name=model_name, model_path=model_path1,
                                               model_type=model_type1,
                                               version_desc=version_desc1, current_stage=version_stage1)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '1')
        self.assertEqual(response.model_path, model_path1)
        self.assertEqual(response.model_type, model_type1)
        self.assertEqual(response.version_desc, version_desc1)
        self.assertEqual(response.current_stage, version_stage1)

        model_path2 = 'fs://source2.pkl'
        model_type2 = '{"flavor.version":2}'
        version_desc2 = 'test update model version2'
        version_stage2 = ModelVersionStage.VALIDATED
        response = client.update_model_version(model_name=model_name, model_version='1',
                                               model_path=model_path2, model_type=model_type2,
                                               version_desc=version_desc2, current_stage=version_stage2)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '1')
        self.assertEqual(response.model_path, model_path2)
        self.assertEqual(response.model_type, model_type2)
        self.assertEqual(response.version_desc, version_desc2)
        self.assertEqual(response.current_stage, version_stage2)

        response = client.update_model_version(model_name=model_name, model_version='1',
                                               current_stage=ModelVersionStage.DEPLOYED)
        self.assertEqual(response.current_stage, ModelVersionStage.DEPLOYED)

    def test_delete_model_version(self):
        model_name = 'test_delete_model_version'
        model_desc = 'test delete model version'
        response = client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)

        model_path = 'fs://source.pkl'
        model_type = '{"flavor.version":1}'
        version_desc = 'test delete model version'
        response = client.create_model_version(model_name=model_name, model_path=model_path,
                                               model_type=model_type, version_desc=version_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '1')
        self.assertEqual(response.model_path, model_path)
        self.assertEqual(response.model_type, model_type)
        self.assertEqual(response.version_desc, version_desc)

        client.delete_model_version(model_name, '1')
        response = client.get_model_version_detail(model_name, '1')
        self.assertIsNone(response)

    def test_get_model_version_detail(self):
        model_name = 'test_get_model_version_detail'
        model_desc = 'test get model version detail'
        response = client.create_registered_model(model_name=model_name, model_desc=model_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)

        model_path = 'fs://source.pkl'
        model_type = '{"flavor.version":1}'
        version_desc = 'test get model version detail'
        response = client.create_model_version(model_name=model_name, model_path=model_path,
                                               model_type=model_type, version_desc=version_desc)
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '1')
        self.assertEqual(response.model_path, model_path)
        self.assertEqual(response.model_type, model_type)
        self.assertEqual(response.version_desc, version_desc)

        response = client.get_model_version_detail(model_name, '1')
        self.assertIsNotNone(response)
        self.assertEqual(response.model_name, model_name)
        self.assertEqual(response.model_version, '1')
        self.assertEqual(response.model_path, model_path)
        self.assertEqual(response.model_type, model_type)
        self.assertEqual(response.version_desc, version_desc)

    def test_update_and_list_notification(self):
        key = 'test_publish_event_key'
        value1 = 'test_publish_event_value1'
        response = client.publish_event(key=key, value=value1)
        self.assertIsNotNone(response)
        self.assertEqual(response.key, key)
        self.assertEqual(response.value, value1)
        self.assertTrue(response.version > 0)
        notifications = client.list_events(key=key)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].key, key)
        self.assertEqual(notifications[0].value, value1)
        self.assertEqual(notifications[0].version, response.version)
        notifications = client.list_events(key=key, version=0)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].key, key)
        self.assertEqual(notifications[0].value, value1)

        value2 = 'test_publish_event_value2'
        old_response = response
        response = client.publish_event(key=key, value=value2)
        self.assertIsNotNone(response)
        self.assertEqual(response.version, old_response.version + 1)
        notifications = client.list_events(key=key)
        self.assertEqual(len(notifications), 2)
        self.assertEqual(notifications[1].key, key)
        self.assertEqual(notifications[1].value, value2)
        self.assertEqual(notifications[1].version, old_response.version + 1)
        notifications = client.list_events(key=key, version=old_response.version)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].key, key)
        self.assertEqual(notifications[0].value, value2)

        old_response = response
        response = client.publish_event(key=key, value=value2)
        self.assertIsNotNone(response)
        self.assertEqual(response.version, old_response.version + 1)
        notifications = client.list_events(key=key)
        self.assertEqual(len(notifications), 3)
        self.assertEqual(notifications[2].key, key)
        self.assertEqual(notifications[2].value, value2)
        self.assertEqual(notifications[2].version, old_response.version + 1)
        notifications = client.list_events(key=key, version=old_response.version)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].key, key)
        self.assertEqual(notifications[0].value, value2)

    def test_listen_notification(self):
        class TestWatcher(EventWatcher):

            def __init__(self, event_type, test_case: TestCase):
                super(TestWatcher, self).__init__()
                self.event_type = event_type
                self.test_case = test_case

            def process(self, notifications):
                self.test_case.assertNotEqual(len(notifications), 0)
                for notification in notifications:
                    print(notification)

        event_type1 = 'test_listen_notification1'
        key1 = 'test_listen_notification_key1'
        client.start_listen_event(key=key1,
                                  watcher=TestWatcher(event_type1, self))
        client.start_listen_event(key=key1,
                                  watcher=TestWatcher(event_type1, self))
        client1.start_listen_event(key=key1,
                                   watcher=TestWatcher(event_type1, self))
        client2.start_listen_event(key=key1,
                                   watcher=TestWatcher(event_type1, self))

        value1 = 'test_listen_notification_value1'
        client.publish_event(key=key1, value=value1)
        value2 = 'test_listen_notification_value2'
        client.publish_event(key=key1, value=value2)

        time.sleep(10)
        value3 = 'test_listen_notification_value3'
        client.publish_event(key=key1, value=value3)

        time.sleep(1)
        client.stop_listen_event(key1)
        client1.stop_listen_event(key1)
        client2.stop_listen_event(key1)

        key2 = 'test_listen_notification_key2'
        client.publish_event(key=key2, value=value1)
        client.publish_event(key=key2, value=value2)

        event_type2 = 'test_listen_notification2'
        client.start_listen_event(key=key2,
                                  watcher=TestWatcher(event_type2, self))
        client1.start_listen_event(key=key2,
                                   watcher=TestWatcher(event_type2, self))
        client2.start_listen_event(key=key2,
                                   watcher=TestWatcher(event_type2, self))

        time.sleep(10)
        client.publish_event(key=key2, value=value3)

        time.sleep(1)
        client.stop_listen_event(key2)
        client1.stop_listen_event(key2)
        client2.stop_listen_event(key2)

    # def test_submit_workflow(self):
    #
    #     def create_job(index) -> BaseJob:
    #         job: BaseJob = LocalCMDJob(exec_cmd='echo "hello {}" && sleep 1'.format(str(index)),
    #                                    job_context=JobContext(),
    #                                    job_config=BaseJobConfig(engine="cmd_line", platform="local"))
    #         job.instance_id = str(index)
    #         return job
    #
    #     def create_workflow() -> Workflow:
    #         ex_workflow = Workflow()
    #         for i in range(3):
    #             job = create_job(i)
    #             ex_workflow.add_job(job)
    #         deps = [JobControlEdge(target_node_id='0', source_node_id='2',
    #                                signal_config=SignalConfig(signal_key=generate_job_status_key('0'),
    #                                                           signal_value=State.FINISHED.value)),
    #                 JobControlEdge(target_node_id='1', source_node_id='2',
    #                                signal_config=SignalConfig(signal_key=generate_job_status_key('1'),
    #                                                           signal_value=State.FINISHED.value))]
    #         ex_workflow.add_edges("2", deps)
    #         workflow_meta = client.register_workflow_execution(name=generate_time_str(),
    #                                                            project_id=None,
    #                                                            execution_state=State.INIT,
    #                                                            workflow_json=dumps(ex_workflow))
    #         ex_workflow.workflow_id = workflow_meta.uuid
    #         return ex_workflow
    #
    #     workflow = create_workflow()
    #     res = client.submit_workflow(json_utils.dumps(workflow))
    #     self.assertEqual(0, res[0])
    #     workflow_id = res[1]
    #     res = client.stop_workflow(workflow_id=workflow_id)
    #     self.assertEqual(0, res[0])
    #     while client.is_alive_workflow(workflow_id)[1]:
    #         time.sleep(1)
    #     self.assertEqual(1, res[0])
    #     execution_meta = client.get_workflow_execution_by_id(workflow_id)
    #     self.assertEqual(State.FINISHED, execution_meta.execution_state)

    def test_dataset_metric_meta(self):
        start_time = round(time.time())
        end_time = start_time + 1
        metric_meta = client.register_metric_meta(metric_name='test_dataset_metric_meta_1',
                                                  metric_type=MetricType.DATASET,
                                                  project_name='test_dataset_metric_meta_project_1',
                                                  dataset_name='test_dataset_metric_meta_dataset_1',
                                                  job_name='test_dataset_metric_meta_job',
                                                  start_time=start_time, end_time=end_time, uri='/tmp/metric',
                                                  tags='test_dataset_metric_meta', properties=Properties({'a': 'a'}))[2]
        metric_meta = client.get_metric_meta(metric_meta.metric_name)[2]
        self.assertEqual('test_dataset_metric_meta_1', metric_meta.metric_name)
        self.assertEqual(MetricType.DATASET, MetricType.value_of(metric_meta.metric_type))
        self.assertEqual('test_dataset_metric_meta_project_1', metric_meta.project_name)
        self.assertEqual('test_dataset_metric_meta_dataset_1', metric_meta.dataset_name)
        self.assertEqual(start_time, metric_meta.start_time)
        self.assertEqual(end_time, metric_meta.end_time)
        self.assertEqual('/tmp/metric', metric_meta.uri)
        self.assertEqual('test_dataset_metric_meta', metric_meta.tags)
        self.assertEqual(metric_meta.properties['a'], metric_meta.properties['a'])
        metric_meta = client.update_metric_meta(metric_name=metric_meta.metric_name,
                                                dataset_name='test_dataset_metric_meta_dataset_2')[2]
        metric_meta = client.get_metric_meta(metric_meta.metric_name)[2]
        self.assertEqual('test_dataset_metric_meta_dataset_2', metric_meta.dataset_name)
        metric_meta = client.register_metric_meta(metric_name='test_dataset_metric_meta_2',
                                                  metric_type=metric_meta.metric_type,
                                                  project_name='test_dataset_metric_meta_project_2',
                                                  dataset_name=metric_meta.dataset_name,
                                                  job_name=metric_meta.job_name,
                                                  start_time=metric_meta.start_time, end_time=metric_meta.end_time,
                                                  uri=metric_meta.uri,
                                                  tags=metric_meta.tags, properties=metric_meta.properties)[2]
        metric_metas = client.list_dataset_metric_metas(dataset_name=metric_meta.dataset_name)[2]
        self.assertEqual(2, len(metric_metas))
        metric_meta = client.list_dataset_metric_metas(dataset_name=metric_meta.dataset_name,
                                                       project_name=metric_meta.project_name)[2]
        self.assertEqual('test_dataset_metric_meta_2', metric_meta.metric_name)
        self.assertEqual('test_dataset_metric_meta_project_2', metric_meta.project_name)
        self.assertTrue(client.delete_metric_meta(metric_name=metric_meta.metric_name))
        metric_metas = client.list_dataset_metric_metas(dataset_name=metric_meta.dataset_name)[2]
        self.assertTrue(isinstance(metric_metas, MetricMeta))
        self.assertEqual('test_dataset_metric_meta_1', metric_metas.metric_name)

    def test_model_metric_meta(self):
        metric_meta = client.register_metric_meta(metric_name='test_model_metric_meta_1',
                                                  metric_type=MetricType.MODEL,
                                                  project_name='test_model_metric_meta_project_1',
                                                  model_name='test_model_metric_meta_model_1',
                                                  job_name='test_model_metric_meta_job',
                                                  uri='/tmp/metric',
                                                  tags='test_model_metric_meta', properties=Properties({'a': 'a'}))[2]
        metric_meta = client.get_metric_meta(metric_meta.metric_name)[2]
        self.assertEqual('test_model_metric_meta_1', metric_meta.metric_name)
        self.assertEqual(MetricType.MODEL, MetricType.value_of(metric_meta.metric_type))
        self.assertEqual('test_model_metric_meta_project_1', metric_meta.project_name)
        self.assertEqual('test_model_metric_meta_model_1', metric_meta.model_name)
        self.assertEqual('/tmp/metric', metric_meta.uri)
        self.assertEqual('test_model_metric_meta', metric_meta.tags)
        self.assertEqual(metric_meta.properties['a'], metric_meta.properties['a'])
        metric_meta = client.update_metric_meta(metric_name=metric_meta.metric_name,
                                                model_name='test_model_metric_meta_model_2')[2]
        metric_meta = client.get_metric_meta(metric_meta.metric_name)[2]
        self.assertEqual('test_model_metric_meta_model_2', metric_meta.model_name)
        metric_meta = client.register_metric_meta(metric_name='test_model_metric_meta_2',
                                                  metric_type=metric_meta.metric_type,
                                                  project_name='test_model_metric_meta_project_2',
                                                  model_name=metric_meta.model_name,
                                                  job_name=metric_meta.job_name,
                                                  uri=metric_meta.uri,
                                                  tags=metric_meta.tags, properties=metric_meta.properties)[2]
        metric_metas = client.list_model_metric_metas(model_name=metric_meta.model_name)[2]
        self.assertEqual(2, len(metric_metas))
        metric_meta = client.list_model_metric_metas(model_name=metric_meta.model_name,
                                                     project_name=metric_meta.project_name)[2]
        self.assertEqual('test_model_metric_meta_2', metric_meta.metric_name)
        self.assertEqual('test_model_metric_meta_project_2', metric_meta.project_name)
        self.assertTrue(client.delete_metric_meta(metric_name=metric_meta.metric_name))
        metric_metas = client.list_model_metric_metas(model_name=metric_meta.model_name)[2]
        self.assertTrue(isinstance(metric_metas, MetricMeta))
        self.assertEqual('test_model_metric_meta_1', metric_metas.metric_name)

    def test_metric_summary(self):
        metric_timestamp = round(time.time())
        metric_summary = client.register_metric_summary(metric_name='test_metric_summary_1', metric_key='auc',
                                                        metric_value='0.6', metric_timestamp=metric_timestamp)[2]
        metric_summary = client.get_metric_summary(metric_summary.uuid)[2]
        self.assertEqual(1, metric_summary.uuid)
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.6', metric_summary.metric_value)
        self.assertEqual(metric_timestamp, metric_summary.metric_timestamp)
        metric_summary = client.update_metric_summary(uuid=metric_summary.uuid, metric_value='0.8')[2]
        metric_summary = client.get_metric_summary(metric_summary.uuid)[2]
        self.assertEqual('0.8', metric_summary.metric_value)
        metric_summary = client.register_metric_summary(metric_name=metric_summary.metric_name,
                                                        metric_key=metric_summary.metric_key,
                                                        metric_value='0.7', metric_timestamp=metric_timestamp + 1,
                                                        model_version='test_metric_summary_model_version_1')[2]
        metric_summary = client.register_metric_summary(metric_name=metric_summary.metric_name,
                                                        metric_key='roc',
                                                        metric_value='0.9', metric_timestamp=metric_timestamp + 1,
                                                        model_version='test_metric_summary_model_version_2')[2]
        metric_summaries = client.list_metric_summaries(metric_name=metric_summary.metric_name)[2]
        self.assertEqual(3, len(metric_summaries))
        self.assertEqual('auc', metric_summaries[0].metric_key)
        self.assertEqual('0.8', metric_summaries[0].metric_value)
        self.assertEqual('auc', metric_summaries[1].metric_key)
        self.assertEqual('0.7', metric_summaries[1].metric_value)
        self.assertEqual('roc', metric_summaries[2].metric_key)
        self.assertEqual('0.9', metric_summaries[2].metric_value)
        metric_summaries = client.list_metric_summaries(metric_key='auc')[2]
        self.assertEqual(2, len(metric_summaries))
        self.assertEqual('0.8', metric_summaries[0].metric_value)
        self.assertEqual('0.7', metric_summaries[1].metric_value)
        metric_summary = client.list_metric_summaries(model_version='test_metric_summary_model_version_1')[2]
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.7', metric_summary.metric_value)
        metric_summary = client.list_metric_summaries(model_version='test_metric_summary_model_version_1')[2]
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.7', metric_summary.metric_value)
        metric_summaries = client.list_metric_summaries(metric_name=metric_summary.metric_name,
                                                        start_time=metric_timestamp + 1,
                                                        end_time=metric_summary.metric_timestamp)[2]
        self.assertEqual(2, len(metric_summaries))
        self.assertEqual('auc', metric_summaries[0].metric_key)
        self.assertEqual('0.7', metric_summaries[0].metric_value)
        self.assertEqual('roc', metric_summaries[1].metric_key)
        self.assertEqual('0.9', metric_summaries[1].metric_value)
        metric_summary = client.list_metric_summaries(metric_name=metric_summary.metric_name, metric_key='auc',
                                                      model_version='test_metric_summary_model_version_1')[2]
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.7', metric_summary.metric_value)

    @staticmethod
    def register_workflow_job():
        project = client.register_project(name='project')

        return project

    @staticmethod
    def register_model_and_version(project):
        model_name = 'test_create_registered_model'
        model_desc = 'test create registered model'
        model = client.register_model(model_name=model_name, project_id=project.uuid,
                                      model_desc=model_desc)
        version = client.register_model_version(model=model.uuid,
                                                model_path="/tmp",
                                                project_snapshot_id=None)
        return model, version


class TestAIFlowClientSqlite(AIFlowClientTestCases, unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        global client, client1, client2
        print("TestAIFlowClientSqlite setUpClass")
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        cls.server = AIFlowServer(store_uri=_SQLITE_DB_URI, port=_PORT, start_scheduler_service=False)
        cls.server.run()
        client = AIFlowClient(server_uri='localhost:' + _PORT)
        client1 = AIFlowClient(server_uri='localhost:' + _PORT)
        client2 = AIFlowClient(server_uri='localhost:' + _PORT)

    @classmethod
    def tearDownClass(cls) -> None:
        client.stop_listen_event()
        client1.stop_listen_event()
        client2.stop_listen_event()
        cls.server.stop()
        os.remove(_SQLITE_DB_FILE)

    def setUp(self) -> None:
        _get_store(_SQLITE_DB_URI)

    def tearDown(self) -> None:
        store = _get_store(_SQLITE_DB_URI)
        base.metadata.drop_all(store.db_engine)


class TestAIFlowClientSqliteWithSingleHighAvailableServer(
        AIFlowClientTestCases, unittest.TestCase):
    """
    Used to ensure the high available server has the same functionality with normal server.
    """

    @classmethod
    def setUpClass(cls) -> None:
        global client, client1, client2
        print("TestAIFlowClientSqlite setUpClass")
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        cls.server = AIFlowServer(store_uri=_SQLITE_DB_URI, port=_PORT, enabled_ha=True, start_scheduler_service=False,
                                  ha_server_uri='localhost:' + _PORT)
        cls.server.run()
        config = ProjectConfig()
        config.set_server_ip('localhost')
        config.set_server_port('50051')
        config.set_project_name('test_project')
        config.set_enable_ha(True)
        client = AIFlowClient(server_uri='localhost:' + _PORT, project_config=config)
        client1 = AIFlowClient(server_uri='localhost:' + _PORT, project_config=config)
        client2 = AIFlowClient(server_uri='localhost:' + _PORT, project_config=config)

    @classmethod
    def tearDownClass(cls) -> None:
        client.stop_listen_event()
        client.disable_high_availability()
        client1.stop_listen_event()
        client1.disable_high_availability()
        client2.stop_listen_event()
        client2.disable_high_availability()
        cls.server.stop()
        os.remove(_SQLITE_DB_FILE)

    def setUp(self) -> None:
        _get_store(_SQLITE_DB_URI)

    def tearDown(self) -> None:
        store = _get_store(_SQLITE_DB_URI)
        base.metadata.drop_all(store.db_engine)


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Dict, List, Optional, Set, Union

from airflow.configuration import conf
from airflow.executors.base_executor import CommandType, EventBufferValueType, QueuedTaskInstanceType
from airflow.executors.celery_executor import CeleryExecutor
from airflow.executors.kubernetes_executor import KubernetesExecutor
from airflow.models.taskinstance import SimpleTaskInstance, TaskInstance, TaskInstanceKey
from airflow.utils.log.logging_mixin import LoggingMixin


class CeleryKubernetesExecutor(LoggingMixin):
    """
    CeleryKubernetesExecutor consists of CeleryExecutor and KubernetesExecutor.
    It chooses an executor to use based on the queue defined on the task.
    When the queue is `kubernetes`, KubernetesExecutor is selected to run the task,
    otherwise, CeleryExecutor is used.
    """

    KUBERNETES_QUEUE = conf.get('celery_kubernetes_executor', 'kubernetes_queue')

    def __init__(self, celery_executor, kubernetes_executor):
        super().__init__()
        self.celery_executor = celery_executor
        self.kubernetes_executor = kubernetes_executor

    @property
    def queued_tasks(self) -> Dict[TaskInstanceKey, QueuedTaskInstanceType]:
        """Return queued tasks from celery and kubernetes executor"""
        queued_tasks = self.celery_executor.queued_tasks.copy()
        queued_tasks.update(self.kubernetes_executor.queued_tasks)

        return queued_tasks

    @property
    def running(self) -> Set[TaskInstanceKey]:
        """Return running tasks from celery and kubernetes executor"""
        return self.celery_executor.running.union(self.kubernetes_executor.running)

    def start(self) -> None:
        """Start celery and kubernetes executor"""
        self.celery_executor.start()
        self.kubernetes_executor.start()

    def queue_command(
        self,
        task_instance: TaskInstance,
        command: CommandType,
        priority: int = 1,
        queue: Optional[str] = None,
    ):
        """Queues command via celery or kubernetes executor"""
        executor = self._router(task_instance)
        self.log.debug("Using executor: %s for %s", executor.__class__.__name__, task_instance.key)
        executor.queue_command(task_instance, command, priority, queue)

    def queue_task_instance(
        self,
        task_instance: TaskInstance,
        mark_success: bool = False,
        pickle_id: Optional[str] = None,
        ignore_all_deps: bool = False,
        ignore_depends_on_past: bool = False,
        ignore_task_deps: bool = False,
        ignore_ti_state: bool = False,
        pool: Optional[str] = None,
        cfg_path: Optional[str] = None,
    ) -> None:
        """Queues task instance via celery or kubernetes executor"""
        executor = self._router(SimpleTaskInstance(task_instance))
        self.log.debug(
            "Using executor: %s to queue_task_instance for %s", executor.__class__.__name__, task_instance.key
        )
        executor.queue_task_instance(
            task_instance,
            mark_success,
            pickle_id,
            ignore_all_deps,
            ignore_depends_on_past,
            ignore_task_deps,
            ignore_ti_state,
            pool,
            cfg_path,
        )

    def has_task(self, task_instance: TaskInstance) -> bool:
        """
        Checks if a task is either queued or running in either celery or kubernetes executor.

        :param task_instance: TaskInstance
        :return: True if the task is known to this executor
        """
        return self.celery_executor.has_task(task_instance) or self.kubernetes_executor.has_task(
            task_instance
        )

    def heartbeat(self) -> None:
        """Heartbeat sent to trigger new jobs in celery and kubernetes executor"""
        self.celery_executor.heartbeat()
        self.kubernetes_executor.heartbeat()

    def get_event_buffer(self, dag_ids=None) -> Dict[TaskInstanceKey, EventBufferValueType]:
        """
        Returns and flush the event buffer from celery and kubernetes executor

        :param dag_ids: to dag_ids to return events for, if None returns all
        :return: a dict of events
        """
        cleared_events_from_celery = self.celery_executor.get_event_buffer(dag_ids)
        cleared_events_from_kubernetes = self.kubernetes_executor.get_event_buffer(dag_ids)

        return {**cleared_events_from_celery, **cleared_events_from_kubernetes}

    def try_adopt_task_instances(self, tis: List[TaskInstance]) -> List[TaskInstance]:
        """
        Try to adopt running task instances that have been abandoned by a SchedulerJob dying.

        Anything that is not adopted will be cleared by the scheduler (and then become eligible for
        re-scheduling)

        :return: any TaskInstances that were unable to be adopted
        :rtype: list[airflow.models.TaskInstance]
        """
        celery_tis = []
        kubernetes_tis = []
        abandoned_tis = []
        for ti in tis:
            if ti.queue == self.KUBERNETES_QUEUE:
                kubernetes_tis.append(ti)
            else:
                celery_tis.append(ti)
        abandoned_tis.extend(self.celery_executor.try_adopt_task_instances(celery_tis))
        abandoned_tis.extend(self.kubernetes_executor.try_adopt_task_instances(kubernetes_tis))
        return abandoned_tis

    def end(self) -> None:
        """End celery and kubernetes executor"""
        self.celery_executor.end()
        self.kubernetes_executor.end()

    def terminate(self) -> None:
        """Terminate celery and kubernetes executor"""
        self.celery_executor.terminate()
        self.kubernetes_executor.terminate()

    def _router(self, simple_task_instance: SimpleTaskInstance) -> Union[CeleryExecutor, KubernetesExecutor]:
        """
        Return either celery_executor or kubernetes_executor

        :param simple_task_instance: SimpleTaskInstance
        :return: celery_executor or kubernetes_executor
        :rtype: Union[CeleryExecutor, KubernetesExecutor]
        """
        if simple_task_instance.queue == self.KUBERNETES_QUEUE:
            return self.kubernetes_executor
        return self.celery_executor

    def debug_dump(self):
        """Called in response to SIGUSR2 by the scheduler"""
        self.log.info("Dumping CeleryExecutor state")
        self.celery_executor.debug_dump()
        self.log.info("Dumping KubernetesExecutor state")
        self.kubernetes_executor.debug_dump()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Serialized DAG and BaseOperator"""
import datetime
import enum
import logging
from inspect import Parameter, signature
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Union

import cattr
import pendulum
from dateutil import relativedelta

try:
    from functools import cache
except ImportError:
    from functools import lru_cache

    cache = lru_cache(maxsize=None)
from pendulum.tz.timezone import Timezone

from airflow.exceptions import AirflowException, SerializationError
from airflow.models.baseoperator import BaseOperator, BaseOperatorLink
from airflow.models.eventhandler import EventHandler
from airflow.models.connection import Connection
from airflow.models.dag import DAG
from airflow.providers_manager import ProvidersManager
from airflow.serialization.enums import DagAttributeTypes as DAT, Encoding
from airflow.serialization.helpers import serialize_template_field
from airflow.serialization.json_schema import Validator, load_dag_schema
from airflow.settings import json
from airflow.utils.code_utils import get_python_source
from airflow.utils.module_loading import import_string
from airflow.utils.task_group import TaskGroup

try:
    # isort: off
    from kubernetes.client import models as k8s
    from airflow.kubernetes.pod_generator import PodGenerator

    # isort: on
    HAS_KUBERNETES = True
except ImportError:
    HAS_KUBERNETES = False


if TYPE_CHECKING:
    from airflow.ti_deps.deps.base_ti_dep import BaseTIDep


log = logging.getLogger(__name__)

_OPERATOR_EXTRA_LINKS: Set[str] = {
    "airflow.operators.trigger_dagrun.TriggerDagRunLink",
    "airflow.sensors.external_task.ExternalTaskSensorLink",
    # Deprecated names, so that existing serialized dags load straight away.
    "airflow.operators.dagrun_operator.TriggerDagRunLink",
    "airflow.sensors.external_task_sensor.ExternalTaskSensorLink",
}


@cache
def get_operator_extra_links():
    """
    Returns operator extra links - both the ones that are built in and the ones that come from
    the providers.

    :return: set of extra links
    """
    _OPERATOR_EXTRA_LINKS.update(ProvidersManager().extra_links_class_names)
    return _OPERATOR_EXTRA_LINKS


class BaseSerialization:
    """BaseSerialization provides utils for serialization."""

    # JSON primitive types.
    _primitive_types = (int, bool, float, str)

    # Time types.
    # datetime.date and datetime.time are converted to strings.
    _datetime_types = (datetime.datetime,)

    # Object types that are always excluded in serialization.
    _excluded_types = (logging.Logger, Connection, type)

    _json_schema: Optional[Validator] = None

    _CONSTRUCTOR_PARAMS: Dict[str, Parameter] = {}

    SERIALIZER_VERSION = 1

    @classmethod
    def to_json(cls, var: Union[DAG, BaseOperator, dict, list, set, tuple]) -> str:
        """Stringifies DAGs and operators contained by var and returns a JSON string of var."""
        return json.dumps(cls.to_dict(var), ensure_ascii=True)

    @classmethod
    def to_dict(cls, var: Union[DAG, BaseOperator, dict, list, set, tuple]) -> dict:
        """Stringifies DAGs and operators contained by var and returns a dict of var."""
        # Don't call on this class directly - only SerializedDAG or
        # SerializedBaseOperator should be used as the "entrypoint"
        raise NotImplementedError()

    @classmethod
    def from_json(cls, serialized_obj: str) -> Union['BaseSerialization', dict, list, set, tuple]:
        """Deserializes json_str and reconstructs all DAGs and operators it contains."""
        return cls.from_dict(json.loads(serialized_obj))

    @classmethod
    def from_dict(
        cls, serialized_obj: Dict[Encoding, Any]
    ) -> Union['BaseSerialization', dict, list, set, tuple]:
        """Deserializes a python dict stored with type decorators and
        reconstructs all DAGs and operators it contains.
        """
        return cls._deserialize(serialized_obj)

    @classmethod
    def validate_schema(cls, serialized_obj: Union[str, dict]) -> None:
        """Validate serialized_obj satisfies JSON schema."""
        if cls._json_schema is None:
            raise AirflowException(f'JSON schema of {cls.__name__:s} is not set.')

        if isinstance(serialized_obj, dict):
            cls._json_schema.validate(serialized_obj)
        elif isinstance(serialized_obj, str):
            cls._json_schema.validate(json.loads(serialized_obj))
        else:
            raise TypeError("Invalid type: Only dict and str are supported.")

    @staticmethod
    def _encode(x: Any, type_: Any) -> Dict[Encoding, Any]:
        """Encode data by a JSON dict."""
        return {Encoding.VAR: x, Encoding.TYPE: type_}

    @classmethod
    def _is_primitive(cls, var: Any) -> bool:
        """Primitive types."""
        return var is None or isinstance(var, cls._primitive_types)

    @classmethod
    def _is_excluded(cls, var: Any, attrname: str, instance: Any) -> bool:
        """Types excluded from serialization."""
        if var is None:
            if not cls._is_constructor_param(attrname, instance):
                # Any instance attribute, that is not a constructor argument, we exclude None as the default
                return True

            return cls._value_is_hardcoded_default(attrname, var, instance)
        return isinstance(var, cls._excluded_types) or cls._value_is_hardcoded_default(
            attrname, var, instance
        )

    @classmethod
    def serialize_to_json(
        cls, object_to_serialize: Union[BaseOperator, DAG], decorated_fields: Set
    ) -> Dict[str, Any]:
        """Serializes an object to json"""
        serialized_object: Dict[str, Any] = {}
        keys_to_serialize = object_to_serialize.get_serialized_fields()
        for key in keys_to_serialize:
            # None is ignored in serialized form and is added back in deserialization.
            value = getattr(object_to_serialize, key, None)
            if cls._is_excluded(value, key, object_to_serialize):
                continue

            if key in decorated_fields:
                serialized_object[key] = cls._serialize(value)
            else:
                value = cls._serialize(value)
                if isinstance(value, dict) and "__type" in value:
                    value = value["__var"]
                serialized_object[key] = value
        return serialized_object

    # pylint: disable=too-many-return-statements
    @classmethod
    def _serialize(cls, var: Any) -> Any:  # Unfortunately there is no support for recursive types in mypy
        """Helper function of depth first search for serialization.

        The serialization protocol is:

        (1) keeping JSON supported types: primitives, dict, list;
        (2) encoding other types as ``{TYPE: 'foo', VAR: 'bar'}``, the deserialization
            step decode VAR according to TYPE;
        (3) Operator has a special field CLASS to record the original class
            name for displaying in UI.
        """
        if cls._is_primitive(var):
            # enum.IntEnum is an int instance, it causes json dumps error so we use its value.
            if isinstance(var, enum.Enum):
                return var.value
            return var
        elif isinstance(var, dict):
            return cls._encode({str(k): cls._serialize(v) for k, v in var.items()}, type_=DAT.DICT)
        elif isinstance(var, list):
            return [cls._serialize(v) for v in var]
        elif HAS_KUBERNETES and isinstance(var, k8s.V1Pod):
            json_pod = PodGenerator.serialize_pod(var)
            return cls._encode(json_pod, type_=DAT.POD)
        elif isinstance(var, DAG):
            return SerializedDAG.serialize_dag(var)
        elif isinstance(var, BaseOperator):
            return SerializedBaseOperator.serialize_operator(var)
        elif isinstance(var, cls._datetime_types):
            return cls._encode(var.timestamp(), type_=DAT.DATETIME)
        elif isinstance(var, datetime.timedelta):
            return cls._encode(var.total_seconds(), type_=DAT.TIMEDELTA)
        elif isinstance(var, Timezone):
            return cls._encode(str(var.name), type_=DAT.TIMEZONE)
        elif isinstance(var, relativedelta.relativedelta):
            encoded = {k: v for k, v in var.__dict__.items() if not k.startswith("_") and v}
            if var.weekday and var.weekday.n:
                # Every n'th Friday for example
                encoded['weekday'] = [var.weekday.weekday, var.weekday.n]
            elif var.weekday:
                encoded['weekday'] = [var.weekday.weekday]
            return cls._encode(encoded, type_=DAT.RELATIVEDELTA)
        elif callable(var):
            return str(get_python_source(var))
        elif isinstance(var, set):
            # FIXME: casts set to list in customized serialization in future.
            return cls._encode([cls._serialize(v) for v in var], type_=DAT.SET)
        elif isinstance(var, tuple):
            # FIXME: casts tuple to list in customized serialization in future.
            return cls._encode([cls._serialize(v) for v in var], type_=DAT.TUPLE)
        elif isinstance(var, TaskGroup):
            return SerializedTaskGroup.serialize_task_group(var)
        else:
            log.debug('Cast type %s to str in serialization.', type(var))
            return str(var)

    # pylint: enable=too-many-return-statements

    @classmethod
    def _deserialize(cls, encoded_var: Any) -> Any:  # pylint: disable=too-many-return-statements
        """Helper function of depth first search for deserialization."""
        # JSON primitives (except for dict) are not encoded.
        if cls._is_primitive(encoded_var):
            return encoded_var
        elif isinstance(encoded_var, list):
            return [cls._deserialize(v) for v in encoded_var]

        if not isinstance(encoded_var, dict):
            raise ValueError(f"The encoded_var should be dict and is {type(encoded_var)}")
        var = encoded_var[Encoding.VAR]
        type_ = encoded_var[Encoding.TYPE]

        if type_ == DAT.DICT:
            return {k: cls._deserialize(v) for k, v in var.items()}
        elif type_ == DAT.DAG:
            return SerializedDAG.deserialize_dag(var)
        elif type_ == DAT.OP:
            return SerializedBaseOperator.deserialize_operator(var)
        elif type_ == DAT.DATETIME:
            return pendulum.from_timestamp(var)
        elif type_ == DAT.POD:
            if not HAS_KUBERNETES:
                raise RuntimeError("Cannot deserialize POD objects without kubernetes libraries installed!")
            pod = PodGenerator.deserialize_model_dict(var)
            return pod
        elif type_ == DAT.TIMEDELTA:
            return datetime.timedelta(seconds=var)
        elif type_ == DAT.TIMEZONE:
            return Timezone(var)
        elif type_ == DAT.RELATIVEDELTA:
            if 'weekday' in var:
                var['weekday'] = relativedelta.weekday(*var['weekday'])  # type: ignore
            return relativedelta.relativedelta(**var)
        elif type_ == DAT.SET:
            return {cls._deserialize(v) for v in var}
        elif type_ == DAT.TUPLE:
            return tuple([cls._deserialize(v) for v in var])
        else:
            raise TypeError(f'Invalid type {type_!s} in deserialization.')

    _deserialize_datetime = pendulum.from_timestamp
    _deserialize_timezone = pendulum.tz.timezone

    @classmethod
    def _deserialize_timedelta(cls, seconds: int) -> datetime.timedelta:
        return datetime.timedelta(seconds=seconds)

    @classmethod
    def _is_constructor_param(cls, attrname: str, instance: Any) -> bool:
        # pylint: disable=unused-argument
        return attrname in cls._CONSTRUCTOR_PARAMS

    @classmethod
    def _value_is_hardcoded_default(cls, attrname: str, value: Any, instance: Any) -> bool:
        """
        Return true if ``value`` is the hard-coded default for the given attribute.

        This takes in to account cases where the ``concurrency`` parameter is
        stored in the ``_concurrency`` attribute.

        And by using `is` here only and not `==` this copes with the case a
        user explicitly specifies an attribute with the same "value" as the
        default. (This is because ``"default" is "default"`` will be False as
        they are different strings with the same characters.)

        Also returns True if the value is an empty list or empty dict. This is done
        to account for the case where the default value of the field is None but has the
        ``field = field or {}`` set.
        """
        # pylint: disable=unused-argument
        if attrname in cls._CONSTRUCTOR_PARAMS and (
            cls._CONSTRUCTOR_PARAMS[attrname] is value or (value in [{}, []])
        ):
            return True
        return False


class SerializedBaseOperator(BaseOperator, BaseSerialization):
    """A JSON serializable representation of operator.

    All operators are casted to SerializedBaseOperator after deserialization.
    Class specific attributes used by UI are move to object attributes.
    """

    _decorated_fields = {'executor_config'}

    _CONSTRUCTOR_PARAMS = {
        k: v.default
        for k, v in signature(BaseOperator.__init__).parameters.items()
        if v.default is not v.empty
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # task_type is used by UI to display the correct class type, because UI only
        # receives BaseOperator from deserialized DAGs.
        self._task_type = 'BaseOperator'
        # Move class attributes into object attributes.
        self.ui_color = BaseOperator.ui_color
        self.ui_fgcolor = BaseOperator.ui_fgcolor
        self.template_fields = BaseOperator.template_fields
        self.operator_extra_links = BaseOperator.operator_extra_links

    @property
    def task_type(self) -> str:
        # Overwrites task_type of BaseOperator to use _task_type instead of
        # __class__.__name__.
        return self._task_type

    @task_type.setter
    def task_type(self, task_type: str):
        self._task_type = task_type

    @classmethod
    def serialize_operator(cls, op: BaseOperator) -> Dict[str, Any]:
        """Serializes operator into a JSON object."""
        serialize_op = cls.serialize_to_json(op, cls._decorated_fields)
        serialize_op['_task_type'] = op.__class__.__name__
        serialize_op['_task_module'] = op.__class__.__module__

        # Used to determine if an Operator is inherited from DummyOperator
        serialize_op['_is_dummy'] = op.inherits_from_dummy_operator

        if op.operator_extra_links:
            serialize_op['_operator_extra_links'] = cls._serialize_operator_extra_links(
                op.operator_extra_links
            )

        if op.deps is not BaseOperator.deps:
            # Are the deps different to BaseOperator, if so serialize the class names!
            # For Airflow 2.0 expediency we _only_ allow built in Dep classes.
            # Fix this for 2.0.x or 2.1
            deps = []
            for dep in op.deps:
                klass = type(dep)
                module_name = klass.__module__
                if not module_name.startswith("airflow.ti_deps.deps."):
                    raise SerializationError(
                        f"Cannot serialize {(op.dag.dag_id + '.' + op.task_id)!r} with `deps` from non-core "
                        f"module {module_name!r}"
                    )

                deps.append(f'{module_name}.{klass.__name__}')
            serialize_op['deps'] = deps

        # Store all template_fields as they are if there are JSON Serializable
        # If not, store them as strings
        if op.template_fields:
            for template_field in op.template_fields:
                value = getattr(op, template_field, None)
                if not cls._is_excluded(value, template_field, op):
                    serialize_op[template_field] = serialize_template_field(value)

        event_handler = op.get_events_handler()
        if event_handler is not None:
            serialize_op['_events_handler'] = EventHandler.serialize(op.get_events_handler())

        return serialize_op

    @classmethod
    def deserialize_operator(cls, encoded_op: Dict[str, Any]) -> BaseOperator:
        """Deserializes an operator from a JSON object."""
        from airflow import plugins_manager

        plugins_manager.initialize_extra_operators_links_plugins()

        if plugins_manager.operator_extra_links is None:
            raise AirflowException("Can not load plugins")
        op = SerializedBaseOperator(task_id=encoded_op['task_id'])

        # Extra Operator Links defined in Plugins
        op_extra_links_from_plugin = {}

        if "label" not in encoded_op:
            # Handle deserialization of old data before the introduction of TaskGroup
            encoded_op["label"] = encoded_op["task_id"]

        for ope in plugins_manager.operator_extra_links:
            for operator in ope.operators:
                if (
                    operator.__name__ == encoded_op["_task_type"]
                    and operator.__module__ == encoded_op["_task_module"]
                ):
                    op_extra_links_from_plugin.update({ope.name: ope})

        # If OperatorLinks are defined in Plugins but not in the Operator that is being Serialized
        # set the Operator links attribute
        # The case for "If OperatorLinks are defined in the operator that is being Serialized"
        # is handled in the deserialization loop where it matches k == "_operator_extra_links"
        if op_extra_links_from_plugin and "_operator_extra_links" not in encoded_op:
            setattr(op, "operator_extra_links", list(op_extra_links_from_plugin.values()))

        for k, v in encoded_op.items():

            if k == "_downstream_task_ids":
                v = set(v)
            elif k == "subdag":
                v = SerializedDAG.deserialize_dag(v)
            elif k in {"retry_delay", "execution_timeout"}:
                v = cls._deserialize_timedelta(v)
            elif k in encoded_op["template_fields"]:
                pass
            elif k.endswith("_date"):
                v = cls._deserialize_datetime(v)
            elif k == "_operator_extra_links":
                op_predefined_extra_links = cls._deserialize_operator_extra_links(v)

                # If OperatorLinks with the same name exists, Links via Plugin have higher precedence
                op_predefined_extra_links.update(op_extra_links_from_plugin)

                v = list(op_predefined_extra_links.values())
                k = "operator_extra_links"

            elif k == "deps":
                v = cls._deserialize_deps(v)
            elif k == "_events_handler":
                v = EventHandler.deserialize(encoded_op["_events_handler"])
            elif k in cls._decorated_fields or k not in op.get_serialized_fields():
                v = cls._deserialize(v)
            # else use v as it is

            setattr(op, k, v)

        for k in op.get_serialized_fields() - encoded_op.keys() - cls._CONSTRUCTOR_PARAMS.keys():
            setattr(op, k, None)

        # Set all the template_field to None that were not present in Serialized JSON
        for field in op.template_fields:
            if not hasattr(op, field):
                setattr(op, field, None)

        # Used to determine if an Operator is inherited from DummyOperator
        setattr(op, "_is_dummy", bool(encoded_op.get("_is_dummy", False)))

        return op

    @classmethod
    def _is_excluded(cls, var: Any, attrname: str, op: BaseOperator):
        if var is not None and op.has_dag() and attrname.endswith("_date"):
            # If this date is the same as the matching field in the dag, then
            # don't store it again at the task level.
            dag_date = getattr(op.dag, attrname, None)
            if var is dag_date or var == dag_date:
                return True
        return super()._is_excluded(var, attrname, op)

    @classmethod
    def _deserialize_deps(cls, deps: List[str]) -> Set["BaseTIDep"]:
        instances = set()
        for qualname in set(deps):
            if not qualname.startswith("airflow.ti_deps.deps."):
                log.error("Dep class %r not registered", qualname)
                continue

            try:
                instances.add(import_string(qualname)())
            except ImportError:
                log.warning("Error importing dep %r", qualname, exc_info=True)
        return instances

    @classmethod
    def _deserialize_operator_extra_links(cls, encoded_op_links: list) -> Dict[str, BaseOperatorLink]:
        """
        Deserialize Operator Links if the Classes  are registered in Airflow Plugins.
        Error is raised if the OperatorLink is not found in Plugins too.

        :param encoded_op_links: Serialized Operator Link
        :return: De-Serialized Operator Link
        """
        from airflow import plugins_manager

        plugins_manager.initialize_extra_operators_links_plugins()

        if plugins_manager.registered_operator_link_classes is None:
            raise AirflowException("Can't load plugins")
        op_predefined_extra_links = {}

        for _operator_links_source in encoded_op_links:
            # Get the key, value pair as Tuple where key is OperatorLink ClassName
            # and value is the dictionary containing the arguments passed to the OperatorLink
            #
            # Example of a single iteration:
            #
            #   _operator_links_source =
            #   {
            #       'airflow.providers.google.cloud.operators.bigquery.BigQueryConsoleIndexableLink': {
            #           'index': 0
            #       }
            #   },
            #
            #   list(_operator_links_source.items()) =
            #   [
            #       (
            #           'airflow.providers.google.cloud.operators.bigquery.BigQueryConsoleIndexableLink',
            #           {'index': 0}
            #       )
            #   ]
            #
            #   list(_operator_links_source.items())[0] =
            #   (
            #       'airflow.providers.google.cloud.operators.bigquery.BigQueryConsoleIndexableLink',
            #       {
            #           'index': 0
            #       }
            #   )

            _operator_link_class_path, data = list(_operator_links_source.items())[0]
            if _operator_link_class_path in get_operator_extra_links():
                single_op_link_class = import_string(_operator_link_class_path)
            elif _operator_link_class_path in plugins_manager.registered_operator_link_classes:
                single_op_link_class = plugins_manager.registered_operator_link_classes[
                    _operator_link_class_path
                ]
            else:
                log.error("Operator Link class %r not registered", _operator_link_class_path)
                return {}

            op_predefined_extra_link: BaseOperatorLink = cattr.structure(data, single_op_link_class)

            op_predefined_extra_links.update({op_predefined_extra_link.name: op_predefined_extra_link})

        return op_predefined_extra_links

    @classmethod
    def _serialize_operator_extra_links(cls, operator_extra_links: Iterable[BaseOperatorLink]):
        """
        Serialize Operator Links. Store the import path of the OperatorLink and the arguments
        passed to it. Example
        ``[{'airflow.providers.google.cloud.operators.bigquery.BigQueryConsoleLink': {}}]``

        :param operator_extra_links: Operator Link
        :return: Serialized Operator Link
        """
        serialize_operator_extra_links = []
        for operator_extra_link in operator_extra_links:
            op_link_arguments = cattr.unstructure(operator_extra_link)
            if not isinstance(op_link_arguments, dict):
                op_link_arguments = {}
            serialize_operator_extra_links.append(
                {
                    "{}.{}".format(
                        operator_extra_link.__class__.__module__, operator_extra_link.__class__.__name__
                    ): op_link_arguments
                }
            )

        return serialize_operator_extra_links


class SerializedDAG(DAG, BaseSerialization):
    """
    A JSON serializable representation of DAG.

    A stringified DAG can only be used in the scope of scheduler and webserver, because fields
    that are not serializable, such as functions and customer defined classes, are casted to
    strings.

    Compared with SimpleDAG: SerializedDAG contains all information for webserver.
    Compared with DagPickle: DagPickle contains all information for worker, but some DAGs are
    not pickle-able. SerializedDAG works for all DAGs.
    """

    _decorated_fields = {'schedule_interval', 'default_args', '_access_control'}

    @staticmethod
    def __get_constructor_defaults():  # pylint: disable=no-method-argument
        param_to_attr = {
            'concurrency': '_concurrency',
            'description': '_description',
            'default_view': '_default_view',
            'access_control': '_access_control',
        }
        return {
            param_to_attr.get(k, k): v.default
            for k, v in signature(DAG.__init__).parameters.items()
            if v.default is not v.empty
        }

    _CONSTRUCTOR_PARAMS = __get_constructor_defaults.__func__()  # type: ignore
    del __get_constructor_defaults

    _json_schema = load_dag_schema()

    @classmethod
    def serialize_dag(cls, dag: DAG) -> dict:
        """Serializes a DAG into a JSON object."""
        try:
            serialize_dag = cls.serialize_to_json(dag, cls._decorated_fields)

            serialize_dag["tasks"] = [cls._serialize(task) for _, task in dag.task_dict.items()]
            serialize_dag['_task_group'] = SerializedTaskGroup.serialize_task_group(dag.task_group)
            return serialize_dag
        except SerializationError:
            raise
        except Exception:
            raise SerializationError(f'Failed to serialize dag {dag.dag_id!r}')

    @classmethod
    def deserialize_dag(cls, encoded_dag: Dict[str, Any]) -> 'SerializedDAG':
        """Deserializes a DAG from a JSON object."""
        dag = SerializedDAG(dag_id=encoded_dag['_dag_id'])

        for k, v in encoded_dag.items():
            if k == "_downstream_task_ids":
                v = set(v)
            elif k == "tasks":
                v = {task["task_id"]: SerializedBaseOperator.deserialize_operator(task) for task in v}
                k = "task_dict"
            elif k == "timezone":
                v = cls._deserialize_timezone(v)
            elif k in {"dagrun_timeout"}:
                v = cls._deserialize_timedelta(v)
            elif k.endswith("_date"):
                v = cls._deserialize_datetime(v)
            elif k in cls._decorated_fields:
                v = cls._deserialize(v)
            # else use v as it is

            setattr(dag, k, v)

        # Set _task_group
        # pylint: disable=protected-access
        if "_task_group" in encoded_dag:
            dag._task_group = SerializedTaskGroup.deserialize_task_group(  # type: ignore
                encoded_dag["_task_group"], None, dag.task_dict
            )
        else:
            # This must be old data that had no task_group. Create a root TaskGroup and add
            # all tasks to it.
            dag._task_group = TaskGroup.create_root(dag)
            for task in dag.tasks:
                dag.task_group.add(task)
        # pylint: enable=protected-access

        keys_to_set_none = dag.get_serialized_fields() - encoded_dag.keys() - cls._CONSTRUCTOR_PARAMS.keys()
        for k in keys_to_set_none:
            setattr(dag, k, None)

        setattr(dag, 'full_filepath', dag.fileloc)
        for task in dag.task_dict.values():
            task.dag = dag
            serializable_task: BaseOperator = task

            for date_attr in ["start_date", "end_date"]:
                if getattr(serializable_task, date_attr) is None:
                    setattr(serializable_task, date_attr, getattr(dag, date_attr))

            if serializable_task.subdag is not None:
                setattr(serializable_task.subdag, 'parent_dag', dag)
                serializable_task.subdag.is_subdag = True

            for task_id in serializable_task.downstream_task_ids:
                # Bypass set_upstream etc here - it does more than we want
                # noqa: E501 # pylint: disable=protected-access
                dag.task_dict[task_id]._upstream_task_ids.add(serializable_task.task_id)

        return dag

    @classmethod
    def to_dict(cls, var: Any) -> dict:
        """Stringifies DAGs and operators contained by var and returns a dict of var."""
        json_dict = {"__version": cls.SERIALIZER_VERSION, "dag": cls.serialize_dag(var)}

        # Validate Serialized DAG with Json Schema. Raises Error if it mismatches
        cls.validate_schema(json_dict)
        return json_dict

    @classmethod
    def from_dict(cls, serialized_obj: dict) -> 'SerializedDAG':
        """Deserializes a python dict in to the DAG and operators it contains."""
        ver = serialized_obj.get('__version', '<not present>')
        if ver != cls.SERIALIZER_VERSION:
            raise ValueError(f"Unsure how to deserialize version {ver!r}")
        return cls.deserialize_dag(serialized_obj['dag'])


class SerializedTaskGroup(TaskGroup, BaseSerialization):
    """A JSON serializable representation of TaskGroup."""

    @classmethod
    def serialize_task_group(cls, task_group: TaskGroup) -> Optional[Union[Dict[str, Any]]]:
        """Serializes TaskGroup into a JSON object."""
        if not task_group:
            return None

        serialize_group = {
            "_group_id": task_group._group_id,  # pylint: disable=protected-access
            "prefix_group_id": task_group.prefix_group_id,
            "tooltip": task_group.tooltip,
            "ui_color": task_group.ui_color,
            "ui_fgcolor": task_group.ui_fgcolor,
            "children": {
                label: (DAT.OP, child.task_id)
                if isinstance(child, BaseOperator)
                else (DAT.TASK_GROUP, SerializedTaskGroup.serialize_task_group(child))
                for label, child in task_group.children.items()
            },
            "upstream_group_ids": cls._serialize(list(task_group.upstream_group_ids)),
            "downstream_group_ids": cls._serialize(list(task_group.downstream_group_ids)),
            "upstream_task_ids": cls._serialize(list(task_group.upstream_task_ids)),
            "downstream_task_ids": cls._serialize(list(task_group.downstream_task_ids)),
        }

        return serialize_group

    @classmethod
    def deserialize_task_group(
        cls,
        encoded_group: Dict[str, Any],
        parent_group: Optional[TaskGroup],
        task_dict: Dict[str, BaseOperator],
    ) -> Optional[TaskGroup]:
        """Deserializes a TaskGroup from a JSON object."""
        if not encoded_group:
            return None

        group_id = cls._deserialize(encoded_group["_group_id"])
        kwargs = {
            key: cls._deserialize(encoded_group[key])
            for key in ["prefix_group_id", "tooltip", "ui_color", "ui_fgcolor"]
        }
        group = SerializedTaskGroup(group_id=group_id, parent_group=parent_group, **kwargs)
        group.children = {
            label: task_dict[val]
            if _type == DAT.OP  # type: ignore
            else SerializedTaskGroup.deserialize_task_group(val, group, task_dict)
            for label, (_type, val) in encoded_group["children"].items()
        }
        group.upstream_group_ids = set(cls._deserialize(encoded_group["upstream_group_ids"]))
        group.downstream_group_ids = set(cls._deserialize(encoded_group["downstream_group_ids"]))
        group.upstream_task_ids = set(cls._deserialize(encoded_group["upstream_task_ids"]))
        group.downstream_task_ids = set(cls._deserialize(encoded_group["downstream_task_ids"]))
        return group



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from enum import Enum
from typing import Text, List

from ai_flow.common.properties import Properties
from ai_flow.util.json_utils import Jsonable


class DataType(str, Enum):
    INT32 = 'INT32'
    INT64 = 'INT64'
    FLOAT32 = 'FLOAT32'
    FLOAT64 = 'FLOAT64'
    STRING = 'STRING'
    INT32ARRAY = 'INT32ARRAY'
    INT64ARRAY = 'INT64ARRAY'
    FLOAT32ARRAY = 'FLOAT32ARRAY'
    FLOAT64ARRAY = 'FLOAT64ARRAY'
    STRINGARRAY = 'STRINGARRAY'
    BYTES = 'BYTES'
    BYTESARRAY = 'BYTESARRAY'


class Schema(Jsonable):
    def __init__(self,
                 name_list: List[Text] = None,
                 type_list: List[DataType] = None):
        self.name_list = name_list
        self.type_list = type_list

    def __str__(self):
        return '<\n' \
               'Schema\n' \
               'name_list:{},\n' \
               'type_list:{},\n' \
               '>'.format(self.name_list, self.type_list)


class DatasetMeta(Jsonable):
    """define dataset meta """

    def __init__(self,
                 name: Text,
                 data_format: Text = None,
                 description: Text = None,
                 uri: Text = None,
                 create_time: int = None,
                 update_time: int = None,
                 properties: Properties = None,
                 schema: Schema = None,
                 catalog_name: Text = None,
                 catalog_type: Text = None,
                 catalog_database: Text = None,
                 catalog_connection_uri: Text = None,
                 catalog_table: Text = None,
                 uuid: int = None) -> None:
        """ create dataset meta
        Args:
            name: dataset name
            data_format: csv, json, etc.
            description: dataset description
            uri: data persistent storage
            create_time: create dataset datetime
            update_time: update dataset datetime
            properties: properties for the dataset
            schema: column list
        """
        self.name = name
        self.data_format = data_format
        self.description = description
        self.uri = uri
        self.create_time = create_time
        self.update_time = update_time
        self.properties = properties
        self.schema = schema
        self.catalog_name = catalog_name
        self.catalog_type = catalog_type
        self.catalog_database = catalog_database
        self.catalog_connection_uri = catalog_connection_uri
        self.catalog_table = catalog_table
        self.uuid = uuid

    def __str__(self):
        return '<\n' \
               'DatasetMeta\n' \
               'uuid:{},\n' \
               'name:{},\n' \
               'data_format:{}, \n' \
               'description:{},\n' \
               'uri:{},\n' \
               'create_time:{},\n' \
               'update_time:{},\n' \
               'properties:{},\n' \
               'schema:{},\n' \
               'catalog_name:{},\n'\
               'catalog_type:{},\n' \
               'catalog_database:{},\n' \
               'catalog_connection_uri:{},\n' \
               'catalog_table:{}' \
               '\n>'.format(
            self.uuid, self.name, self.data_format, self.description, self.uri,
            self.create_time, self.update_time, self.properties, self.schema,
            self.catalog_name, self.catalog_type, self.catalog_database,
            self.catalog_connection_uri, self.catalog_table)


# dataset api
def create_dataset(name: Text,
                   data_format: Text = None,
                   description: Text = None,
                   uri: Text = None,
                   create_time: int = None,
                   update_time: int = None,
                   properties: Properties = None,
                   name_list: List[Text] = None,
                   type_list: List[DataType] = None,
                   catalog_name: Text = None,
                   catalog_type: Text = None,
                   catalog_database: Text = None,
                   catalog_connection_uri: Text = None,
                   catalog_table: Text = None,
                   ) -> DatasetMeta:
    schema = Schema(name_list=name_list, type_list=type_list)
    return DatasetMeta(name=name, data_format=data_format, description=description,
                       uri=uri, schema=schema, create_time=create_time, update_time=update_time,
                       properties=properties, catalog_name=catalog_name, catalog_type=catalog_type,
                       catalog_database=catalog_database, catalog_connection_uri=catalog_connection_uri,
                       catalog_table=catalog_table
                       )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.api.ai_flow_context import init_ai_flow_context
from ai_flow.api import workflow_operation
from ai_flow.api.ops import read_dataset, write_dataset, transform, train, predict, evaluate, dataset_validate, \
    model_validate, push_model, user_define_operation, action_on_event, action_on_events, action_on_job_status, \
    action_on_model_version_event, action_on_dataset_event
from ai_flow.context.project_context import current_project_context, current_project_config
from ai_flow.context.workflow_config_loader import current_workflow_config
from ai_flow.context.job_context import job_config
from ai_flow.ai_graph.ai_graph import current_graph
from ai_flow.endpoint.server.server_runner import AIFlowServerRunner
from ai_flow.workflow.control_edge import JobAction, EventCondition, MeetAnyEventCondition, MeetAllEventCondition, \
    ValueCondition, EventLife

# ai_flow.client
from ai_flow.client import *

# ai_flow.common
from ai_flow.common.args import Args, ExecuteArgs
from ai_flow.common.properties import Properties

# ai_flow.meta
from ai_flow.meta import *
from ai_flow.meta.artifact_meta import *
from ai_flow.meta.dataset_meta import *
from ai_flow.meta.job_meta import *
from ai_flow.meta.model_meta import *
from ai_flow.meta.model_meta import *
from ai_flow.meta.project_meta import *



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, Optional

from airflow.exceptions import AirflowException


class DagParam:
    """
    Class that represents a DAG run parameter.

    It can be used to parameterize your dags. You can overwrite its value by setting it on conf
    when you trigger your DagRun.

    This can also be used in templates by accessing {{context.params}} dictionary.

    **Example**:

        with DAG(...) as dag:
          EmailOperator(subject=dag.param('subject', 'Hi from Airflow!'))

    :param current_dag: Dag being used for parameter.
    :type current_dag: airflow.models.DAG
    :param name: key value which is used to set the parameter
    :type name: str
    :param default: Default value used if no parameter was set.
    :type default: Any
    """

    def __init__(self, current_dag, name: str, default: Optional[Any] = None):
        if default:
            current_dag.params[name] = default
        self._name = name
        self._default = default

    def resolve(self, context: Dict) -> Any:
        """Pull DagParam value from DagRun context. This method is run during ``op.execute()``."""
        default = self._default
        if not self._default:
            default = context['params'].get(self._name, None)
        resolved = context['dag_run'].conf.get(self._name, default)
        if not resolved:
            raise AirflowException(f'No value could be resolved for parameter {self._name}')
        return resolved



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""Utilities to dump and load Jsonable object to/from JSONs."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import importlib
import inspect
import json
from typing import Any, Dict, List, Text, Type, Union
import logging
from six import with_metaclass

# This is the special key to indicate the serialized object type.
# Depending on which, the utility knows how to deserialize it back to its
# original type.
_AF_OBJECT_TYPE_KEY = '__af_object_type__'
_MODULE_KEY = '__module__'
_CLASS_KEY = '__class__'
_DATA = '__data__'

_BYTES_ENCODE = "ISO-8859-1"

RUNTIME_PARAMETER_PATTERN = (r'({\\*"__class__\\*": \\*"RuntimeParameter\\*", '
                             r'.*?})')


class _ObjectType(object):
    """Internal class to hold supported types."""
    # Indicates that the JSON dictionary is an instance of Jsonable type.
    # The dictionary has the states of the object and the object type info is
    # stored as __module__ and __class__ fields.
    JSONABLE = 'jsonable'
    # Indicates that the JSON dictionary is a python class.
    # The class info is stored as __module__ and __class__ fields in the
    # dictionary.
    CLASS = 'class'
    # Indicates that the JSON dictionary is an instance of a proto.Message
    # subclass. The class info of the proto python class is stored as __module__
    # and __class__ fields in the dictionary. The serialized value of the proto is
    # stored in the dictionary with key of _PROTO_VALUE_KEY.
    PROTO = 'proto'

    BYTES = 'bytes'

    SET = 'set'


class Jsonable(with_metaclass(abc.ABCMeta, object)):
    """Base class for serializing and deserializing objects to/from JSON.

    The default implementation assumes that the subclass can be restored by
    updating `self.__dict__` without invoking `self.__init__` function.. If the
    subclass cannot hold the assumption, it should
    override `to_json_dict` and `from_json_dict` to customize the implementation.
    """

    def to_json_dict(self) -> Dict[Text, Any]:
        """Convert from an object to a JSON serializable dictionary."""
        return self.__dict__

    @classmethod
    def from_json_dict(cls, dict_data: Dict[Text, Any]) -> Any:
        """Convert from dictionary data to an object."""
        instance = cls.__new__(cls)
        instance.__dict__ = dict_data
        return instance


JsonableValue = Union[bool, bytes, float, int, Jsonable, Text,
                      Type]
JsonableList = List[JsonableValue]
JsonableDict = Dict[Union[bytes, Text], Union[JsonableValue, JsonableList]]
JsonableType = Union[JsonableValue, JsonableList, JsonableDict]


class _DefaultEncoder(json.JSONEncoder):
    """Default JSON Encoder which encodes Jsonable object to JSON."""

    def encode(self, obj: Any) -> Text:
        try:
            """Override encode to prevent redundant dumping."""
            if obj.__class__.__name__ == 'RuntimeParameter' and obj.ptype == Text:
                return self.default(obj)

            return super(_DefaultEncoder, self).encode(obj)
        except Exception as e:
            logging.exception(obj.__class__.__name__)
            raise e

    def default(self, obj: Any) -> Any:
        # If obj is a str-typed RuntimeParameter, serialize it in place.
        if obj.__class__.__name__ == 'RuntimeParameter' and obj.ptype == Text:
            dict_data = {
                _AF_OBJECT_TYPE_KEY: _ObjectType.JSONABLE,
                _MODULE_KEY: obj.__class__.__module__,
                _CLASS_KEY: obj.__class__.__name__,
            }
            dict_data.update(obj.to_json_dict())
            return dumps(dict_data)

        if isinstance(obj, Jsonable):
            dict_data = {
                _AF_OBJECT_TYPE_KEY: _ObjectType.JSONABLE,
                _MODULE_KEY: obj.__class__.__module__,
                _CLASS_KEY: obj.__class__.__name__,
            }
            # Need to first check the existence of str-typed runtime parameter.
            data_patch = obj.to_json_dict()
            for k, v in data_patch.items():
                if v.__class__.__name__ == 'RuntimeParameter' and v.ptype == Text:
                    data_patch[k] = dumps(v)
            dict_data.update(data_patch)
            return dict_data

        if isinstance(obj, bytes):
            dict_data = {
                _AF_OBJECT_TYPE_KEY: _ObjectType.BYTES,
                _MODULE_KEY: obj.__class__.__module__,
                _CLASS_KEY: obj.__class__.__name__,
                _DATA: str(obj, _BYTES_ENCODE)
            }
            return dict_data

        if isinstance(obj, set):
            dict_data = {
                _AF_OBJECT_TYPE_KEY: _ObjectType.SET,
                _MODULE_KEY: obj.__class__.__module__,
                _CLASS_KEY: obj.__class__.__name__,
                _DATA: sorted(list(obj))
            }
            return dict_data

        if inspect.isclass(obj):
            return {
                _AF_OBJECT_TYPE_KEY: _ObjectType.CLASS,
                _MODULE_KEY: obj.__module__,
                _CLASS_KEY: obj.__name__,
            }

        return super(_DefaultEncoder, self).default(obj)


class _DefaultDecoder(json.JSONDecoder):
    """Default JSON Decoder which decodes JSON to Jsonable object."""

    def __init__(self, *args, **kwargs):
        super(_DefaultDecoder, self).__init__(
            object_hook=self._dict_to_object, *args, **kwargs)

    def _dict_to_object(self, dict_data: Dict[Text, Any]) -> Any:
        """Converts a dictionary to an object."""
        if _AF_OBJECT_TYPE_KEY not in dict_data:
            return dict_data

        object_type = dict_data.pop(_AF_OBJECT_TYPE_KEY)

        def _extract_class(d):
            module_name = d.pop(_MODULE_KEY)
            class_name = d.pop(_CLASS_KEY)
            return getattr(importlib.import_module(module_name), class_name)

        if object_type == _ObjectType.JSONABLE:
            jsonable_class_type = _extract_class(dict_data)
            if not issubclass(jsonable_class_type, Jsonable):
                raise ValueError('Class %s must be a subclass of Jsonable' %
                                 jsonable_class_type)
            return jsonable_class_type.from_json_dict(dict_data)

        if object_type == _ObjectType.CLASS:
            return _extract_class(dict_data)

        if object_type == _ObjectType.BYTES:
            return bytes(dict_data.pop(_DATA), _BYTES_ENCODE)

        if object_type == _ObjectType.SET:
            return set(dict_data.pop(_DATA))


def dumps(obj: Any) -> Text:
    """Dumps an object to JSON with Jsonable encoding."""
    return json.dumps(obj, cls=_DefaultEncoder, sort_keys=True)


def loads(s: Text) -> Any:
    """Loads a JSON into an object with Jsonable decoding."""
    return json.loads(s, cls=_DefaultDecoder)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from datetime import datetime
from typing import Optional

from airflow.models.taskinstance import SimpleTaskInstance


class CallbackRequest:
    """
    Base Class with information about the callback to be executed.

    :param full_filepath: File Path to use to run the callback
    :param msg: Additional Message that can be used for logging
    """

    def __init__(self, full_filepath: str, msg: Optional[str] = None):
        self.full_filepath = full_filepath
        self.msg = msg

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)


class TaskCallbackRequest(CallbackRequest):
    """
    A Class with information about the success/failure TI callback to be executed. Currently, only failure
    callbacks (when tasks are externally killed) and Zombies are run via DagFileProcessorProcess.

    :param full_filepath: File Path to use to run the callback
    :param simple_task_instance: Simplified Task Instance representation
    :param is_failure_callback: Flag to determine whether it is a Failure Callback or Success Callback
    :param msg: Additional Message that can be used for logging to determine failure/zombie
    """

    def __init__(
        self,
        full_filepath: str,
        simple_task_instance: SimpleTaskInstance,
        is_failure_callback: Optional[bool] = True,
        msg: Optional[str] = None,
    ):
        super().__init__(full_filepath=full_filepath, msg=msg)
        self.simple_task_instance = simple_task_instance
        self.is_failure_callback = is_failure_callback


class DagCallbackRequest(CallbackRequest):
    """
    A Class with information about the success/failure DAG callback to be executed.

    :param full_filepath: File Path to use to run the callback
    :param dag_id: DAG ID
    :param execution_date: Execution Date for the DagRun
    :param is_failure_callback: Flag to determine whether it is a Failure Callback or Success Callback
    :param msg: Additional Message that can be used for logging
    """

    def __init__(
        self,
        full_filepath: str,
        dag_id: str,
        execution_date: datetime,
        is_failure_callback: Optional[bool] = True,
        msg: Optional[str] = None,
    ):
        super().__init__(full_filepath=full_filepath, msg=msg)
        self.dag_id = dag_id
        self.execution_date = execution_date
        self.is_failure_callback = is_failure_callback


class SlaCallbackRequest(CallbackRequest):
    """
    A class with information about the SLA callback to be executed.

    :param full_filepath: File Path to use to run the callback
    :param dag_id: DAG ID
    """

    def __init__(self, full_filepath: str, dag_id: str):
        super().__init__(full_filepath)
        self.dag_id = dag_id
